function [a, action] = staticact(a, state, actions, rewards, params)
%Implements some static action selection algorithms
%  [A, ACTION] = STATICACT(A, STATE, ACTIONS, REWARDS, PARAMS)
%  Implements a set of static action selection algorithms. They are
%  selected via the following learning parameters:
%   'static.type'  - either one of the following:
%       'fixed'         - a fixed number (the default)
%       'noisy'         - a uniformly distributed random number with the 
%           mean around a given number
%       'rand'          - random choice between discrete actions.
%           Naturally, actions must be discrete
%       'sequence'      - a predefined sequence of actions (e.g., derived
%           from the optimal policy)
%   'static.value'   - the fixed value for 'fixed', mean for 'noisy',
%           the sequence of actions for 'sequence'.
%           Default 0
%   'static.stdev'     - noise standard deviation, for 'noisy'. Default
%           0.1.
%
%
%  See also agent_act

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:

if isfield(params, 'mode') && strcmp(params.mode, 'init'),
    % init defaults
    STATIC.type = 'fixed';
    STATIC.value = 0;
    STATIC.stdev = .1;
    a.learnparam.static = checkparams(a.learnparam.static, STATIC);
    if strcmp(a.learnparam.static.type, 'rand'),
        a.learnparam.static.actionscount = params.info.actionspacesize(params.info.imap(5), 2);
    end;
    if strcmp(a.learnparam.static.type, 'sequence'),
        a.seqindex = 0;
    end;
    return;
end;


s = a.learnparam.static;
switch s.type,
    case 'fixed',
        action = s.value;
        
    case 'noisy',
        action = s.value + randn * s.stdev;
        
    case 'rand',
        action = ceil(rand * s.actionscount);
        
    case 'sequence',
        if params.newtrial,     % reset sequence
            a.seqindex = 0;
        end;
        action = s.value(a.seqindex + 1);
        a.seqindex = mod(a.seqindex + 1, length(s.value)); 
end;

% returning a, action END pigreedyact()
% ==================================================