function [a, action] = teamgreedyact(a, state, actions, rewards, params)
%Implements deterministic greedy action selection for full-state, joint action Q-learners
%  [A, ACTION] = TEAMGREEDYACT(A, STATE, ACTIONS, REWARDS, PARAMS)
%  Implements team-greedy action selection. Supports discrete states and
%  actions, with 1 action variable per agent.
%
%  Values on the agent learning parameters:
%   usesocconv      - whether to use social conventions to break the ties
%       in maximization. Default 1 (use social conventions). The ordering
%       is agents from n to 1, actions from 1 to max.
%
%
%  See also agent_act


%  Author:      Lucian Busoniu
%  Version:     1.1
%  History:
%   - 12 Jan 2006, changed to use monolithic full state instead of just
%   single-agent state, generalized and renamed accordingly
%   - 23 Jun 2006, added support for social conventions

if isfield(params, 'mode') && strcmp(params.mode, 'init'),
    % install default params if necessary
    DEF_LP.usesocconv = 1;
    a.learnparam = checkparams(a.learnparam, DEF_LP);
    return;
end;

ix = ndi2lin([ones(1, length(a.sizes.act)) state'], a.sizes.q);
ix = ix : ix + a.sizes.expact - 1;
topindices = find(a.Q(ix) == max(a.Q(ix)));

% ??? why shouldn't they all be equal (second test)?  I don't know...
if a.learnparam.usesocconv && length(topindices) < a.sizes.expact,
    % instead of breaking ties randomly, choose deterministically the first
    % action in the agent ordering X actions ordering
    % note that due to the special format of the Q-table, actually an
    % inverse agent ordering is used (first agent action varies fastest)
    action = lin2ndi(topindices(1), a.sizes.act);
else
    % find joint action for computed index, breaking ties randomly
    action = lin2ndi(topindices(ceil(rand * length(topindices))), a.sizes.act);
end;


% extract own action
action = action(a.indices(5));

% returning a, action
% END greedyact() =====================================================