function a = asfq(a, state, action, reward, params)
%Implements the adaptive state focus Q-learning algorithm
%  A = ASFQ(A, STATE, ACTION, REWARD, PARAMS)
%  Implements adaptive state focus Q-learning (ASF-Q) [1]. This algorithms
%  works in two modes:
%   - SINGLE, plain Q-learning. However, the evolution of Q-values is
%       monitored and, when lack of convergence is detected, the algorithm
%       switches to mode MULTI.
%   - MULTI, which is full-state Q learning. In this mode, the states of
%       all the agents are considered in maintaining the Q-table.
%
%  Required values on the agent learning parameters:
%   alpha           - the learning rate
%   gamma           - the discount factor
%   lambda          - the eligibility trace decay rate
%   epsilon         - the exploration probability
%   window          - the analysis window length
%   stops           - the number of analysis stops per window
%   zeromargin      - the zero margin, determining when a value used for
%       testing convergence is considered positive, negative or zero
%   expresetweight  - how does the exploration behaviour change upon
%       switching modes
%
%  Required values on the extra parameters:
%   newtrial        - only in episodic environments; whether a new trial is
%                   beginning
%
%  The convergence monitoring happens in the following way: the absolute
%  mean difference between Q-values at successive steps is maintained
%  within a sliding [window] of iterations. The sliding average of this
%  difference is computed in [stops] point per window. After one and a half
%  windows have fully passed, each stop point triggers an analysis. The
%  mean value of the derivative of the sliding average is computed; if this
%  value is not negative, while at the same time the mean value of the
%  sliding average is positive, the algorithm considers itself in a
%  situation of convergence failure. Upon this, it extends the Q-table with
%  the state space of the other agents (switches to mode MULTI). The
%  inequalities above are all done within a robustness margin [zeromargin].
%
%  Supports discrete states and actions, with 1 action variable per agent.
%
%  This learning function should be coupled with the dedicated action
%  function asfq_greedyact().
%
%
%  [1] Busoniu, L., De Schutter, B. and Babuska, R. (2005). Multiagent
%      reinforcement learning with adaptive state focus. In Proceedings 17th
%      Belgian-Dutch Conference on Artificial Intelligence (BNAIC-05), pages
%      35-42, Brussels, Belgium.
%
%  See also agent_learn, asfq_init, asfq_greedyact

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


if a.learnparam.episodic && params.newtrial,
    a.E = 0 * a.Q;          % clear eligibility trace
end;

%%%%%%%%%%%%%%%%%%%%%%%%
%% LEARNING

% determine correct learning indices in Q-table depending on the learning
% mode
switch a.learnmode,
    case 1,         % single
        ix = ndi2lin([1 state(a.indices(1) : a.indices(2))'], a.qsize);
        qix = ndi2lin([action(a.indices(5)) a.state(a.indices(1) : a.indices(2))'], a.qsize);
    case 2,         % multi
        ix = ndi2lin([1 state'], a.qsize);
        qix = ndi2lin([action(a.indices(5)) a.state'], a.qsize);
end;

delta = reward(a.indices(7)) + a.learnparam.gamma * max(a.Q(ix : ix + a.actionscount - 1)) - a.Q(qix);
a.E(qix) = 1;
dQ = a.Q;   % remember previous values on Q table to determine dQ
a.Q = a.Q + a.learnparam.alpha * delta * a.E;
a.E = a.learnparam.gamma * a.learnparam.lambda * a.E * ~a.exploring;

%%%%%%%%%%%%%%%%%%%%%%%%
%% BEST Q CACHE

% an efficiency analysis is performed in SINGLE mode in episodic tasks
% if a trial ends and the currently obtained reward was the greatest,
% presumably the decisions taken along the path were the "most correct";
% hence, the Q-table is stored for possible later retrieval, after the
% coordination problem starts damaging the Q-values

if a.learnmode == 1 && a.learnparam.episodic && a.learnparam.reverttobestq,
    if params.newtrial,
        if isfinite(a.cache.accreward) && a.cache.accreward > a.cache.bestreward,
            a.cache.Q = a.Q;
            a.cache.trial = a.trialsk;
            a.cache.bestreward = a.cache.accreward;
        end;
        a.cache.accreward = 0;
    end;
    a.cache.accreward = a.cache.accreward + reward(a.indices(7));
end;

%%%%%%%%%%%%%%%%%%%%%%%%
%% ANALYSIS

switch a.learnmode,
    case 1,         % single
        % NOTE plug in alternate analysis methods here

        % process Q values variation
        dQ = a.Q - dQ;
        if any(dQ), dQ = mean(abs( dQ(find(dQ)) ));
        else dQ = 0;
        end;
        a.dQ = [a.dQ(2:a.learnparam.window); dQ];
        % perform analysis stop if required
        if a.k > a.learnparam.window && ~mod(a.k, a.learnparam.stoplength),
            a.dQavg = [a.dQavg(2:a.learnparam.stops); mean(a.dQ(find(~isnan(a.dQ))))];
            
%             a.m = [a.m; mean(a.dQ(find(~isnan(a.dQ))))];
%             if ~all(isfinite(a.dQavg)) a.mdot = [a.mdot; NaN]; end;
            
            % if we have enough data, analyse
            if all(isfinite(a.dQavg)),
                % first derivative of sliding mean
                ddQavg = diff(a.dQavg);
                % update reference values (absolute maxima) for sliding mean
                % and its derivative
                a.tempvals.dQavg_ref = max([a.tempvals.dQavg_ref; abs(a.dQavg)]);
                a.tempvals.ddQavg_ref = max([a.tempvals.ddQavg_ref; abs(ddQavg)]);

%                 a.mdot = [a.mdot; mean(ddQavg)];

                % if the derivative is (more or less) positive or zero, and the
                % sliding mean of variation is (more or less) positive, we're
                % in trouble and need to expand the state space
                if (mean(ddQavg) > -a.learnparam.zeromargin * a.tempvals.ddQavg_ref) ...
                        && (mean(a.dQavg) > a.learnparam.zeromargin * a.tempvals.dQavg_ref),

                    % enlarge Q table, retrieving the cache if any
                    if ~isempty(a.cache.Q), a.Q = a.cache.Q; end;
                    smap = a.indices(1):a.indices(2);
                    sz = a.sssize; sz(smap) = 1;
                    % the values get repeated for the rest of the state combinations
                    a.Q = repmat(a.Q, prod(sz), 1);
                    % nevermind actions for a while:
                    % Q has the current agent's state space of the first
                    % positions, and then the state space sizes of other agents
                    % except itself
                    sz = [a.sssize(smap) a.sssize(1 : (a.indices(1)-1)) a.sssize((a.indices(2)+1) : length(a.sssize))];
                    % we need to permute so that this agent's state positions
                    % get back into place, while everything to its left is
                    % shifted into the left side of the sizes
                    perm = [length(smap) + (1 : (a.indices(1)-1)) 1 : length(smap) (a.indices(2)+1) : length(a.sssize)];
                    % TODO these formulae are still doubtful! they work in
                    % the two agent case, but do they for N? need to check

                    % reshape and permute, taking the actions dimension into
                    % account this time, and then flatten back
                    a.Q = permute(reshape(a.Q, [a.actionscount sz]), [1 1+perm]);
                    a.Q = a.Q(:);

                    % the eligibility trace loses its meaning upon Q-table
                    % extension; however, it also gets extended
                    a.E = 0 * a.Q;
                    
                    % log the switch
                    logentry.desc = 'Switching from SINGLE to MULTI mode. Reason: nondecreasing Q-values update sizes.';
%                     disp('switching');
                    if isfield(a.learnparam, 'trialcount'),
                        logentry.trial = a.learnparam.trialcount;
                    end;
                    logentry.iter = a.k;
                    logentry.trial = a.trialsk;
                    if ~isnan(a.cache.trial), logentry.cachetrial = a.cache.trial; end;
                    a.log{length(a.log) + 1} = logentry;
                    
                    % change helper vars
                    a.qsize = [a.actionscount a.sssize];
                    % exploration might get reset
                    if ~isempty(a.learnparam.expresetweight),
                        p.mode = 'reset';
                        p.weight = a.learnparam.expresetweight;
                        a = feval(a.explorefun, a, [], [], [], [], p);
                    end;
                    
                    % switch learning mode and clear previous mode data
                    a.learnmode = 2;
                    rmfield(a, 'cache');

                end;    % expansion if

            end;        % sufficient data for analysis if
        end;            % analysis stop if

%    case 2,        % nothing special happens in multi learning mode
    
end;    % learn mode switch


% returning a
% END plainq() ============================================================
