function a = asfq_init(a, info)
%Initializes the ASF-Q learning algorithm
%  A = ASFQ_INIT(A, INFO)
%  Creates the structures required for the ASF-Q learning algorithm to
%  run. Initializes the Q table and an eligibility trace as flat vectors.
%  Optional values on the agent learning parameters (see asfq() for the
%  list of mandatory values):
%   window          - the analysis window length, default 256. This is
%       the number of iterations over which the differences of Q-values
%       between successive iterations are maintained. For exact analysis
%       stops, make sure this is divisible with 2 and [stops] below.
%   stops           - the number of analysis stops per window, default 16.
%       A sliding mean of the differences is maintained. This mean is
%       subsampled [stops] times per window. The process begins after half
%       of a window has elapsed, thus analysis (which requires a full
%       window of data) can begin after one and a half window.
%   zeromargin      - the zero margin, determining when a value used for
%       testing convergence is considered positive, negative or zero.
%       Positive real between 0 and 1, default 0.05 (5%). A value will be
%       considered positive if it is above [zeromargin]*SignalAmplitude,
%       negative if below -[zeromargin]*SignalAmplitude, zero otherwise.
%   expresetweight  - how does the exploration behaviour change upon
%       switching modes, default []. Either a number between 0 and 1,
%       dictating the exploration reset weight (see agent()), or the empty
%       matrix, meaning switches have no effect on the exploration
%       behaviour.
%   reverttobestq   - if true, the algorithm will maintain a cache of the
%       Q-table that behaved best so far (i.e. accumulated the best reward)
%       and will revert to this Q-table upon expansion. This only works in
%       episodic environments, where "best behaviour so far" makes a clear
%       sense. The default is 0 (don't revert).
%
%  Values on the info parameters:
%   statespacesize      - the state space size
% 
%
%  See also asfq, agent_initlearn

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


% process learning parameters

% Check states and actions
if any(info.statespacesize(:, 1) ~= 1) || ...
        a.indices(5) < a.indices(6) || ...
        info.actionspacesize(a.indices(5), 1) ~= 1
    error('asfq() only supports discrete states and actions, with 1 action variable');
else
    a.actionscount = info.actionspacesize(info.imap(5), 2);
end;

% default values
DEF_PARAMS.window = 256;
DEF_PARAMS.stops = 16;
DEF_PARAMS.zeromargin = 0.05;
DEF_PARAMS.expresetweight = [];
DEF_PARAMS.reverttobestq = 0;

% init to defaults whatever has not been set
pnames = fieldnames(DEF_PARAMS);
for i = 1:length(pnames),
    if ~isfield(a.learnparam, pnames{i}),
        a.learnparam.(pnames{i}) = DEF_PARAMS.(pnames{i});
    end;
end;

% initial learning mode is single
a.learnmode = 1;
% in this mode, we only care about the agent's state, but we save the state
% space size for possible expansion
a.qsize = [a.actionscount info.statespacesize(a.indices(1) : a.indices(2), 2)'];
a.sssize = info.statespacesize(:, 2)';
a.Q = zeros(prod(a.qsize), 1);
a.E = a.Q;

% Q variation
a.dQ = NaN * ones(a.learnparam.window, 1);
% sliding average Q variation
a.learnparam.stoplength = round(a.learnparam.window / a.learnparam.stops);
a.dQavg = NaN * ones(a.learnparam.stops, 1);
a.tempvals.dQavg_ref = NaN;
a.tempvals.ddQavg_ref = NaN;
% best Q cache
a.cache.bestreward = -Inf;  % best accumulated reward so far
a.cache.accreward = -Inf;   % current accumulated reward
a.cache.Q = [];             % current Q cache
a.cache.trial = NaN;        % the trial when the current cache was recorded
% the agent will record switching events in its log
a.log = {};

% a.m = [];
% a.mdot = [];

% declare volatile fields
a.volatile = {'E', 'dQ', 'dQavg', 'tempvals', 'cache'};

% returning a
% END plainq_init() ========================================================