function a = minimaxq2(a, state, action, reward, params)
%Implements an alternate version of the minimax Q-learning algorithm
%  A = MINIMAXQ2(A, STATE, ACTION, REWARD, PARAMS)
%  Implements minimax Q-learning in a faster learning version than Littman's.
%  Discards the auxiliary table for state values used by Littman, computing
%  the next state value at each step by solving an LP problem similar to 
%  the one solved for computing the policy at the current step. Note that
%  this actually means slower running speed, but fewer learning iterations
%  for the same performance. However, the theoretical convergence
%  guarantees are lost in this version.
%
%  Required values on the agent learning parameters:
%   alpha           - the learning rate
%   gamma           - the discount factor
%   lambda          - the eligibility trace decay rate
%   epsilon         - the exploration probability
%  Required values on the extra parameters:
%   newtrial        - only in episodic environments; whether a new trial is
%                   beginning
%
%
%  References:
%  [1] Littman, M. L. (2001). Friend-or-foe Q-learning in general-sum
%      games. In Proceedings of the Eighteenth International Conference on
%      Machine Learning (ICML-01), pages 322-328, Williams College,
%      Williamstown Massachusets, USA.
%
%  See agent_learn, minimaxq2_init


% by convention, all idle agents "take" deterministically their
% first action, so that learning is simplified for other agents
action(find(~action)) = 1;

if a.learnparam.episodic && params.newtrial, 
    a.E = 0 * a.Q;          % clear eligibility trace
end;

% ==========================================
% == do the TD(lambda) update

% compute V(s')
vkplus1ix = ndi2lin([ones(1, length(action)) state'], a.sizes.q);
Qkplus1 = reshape(a.Q(vkplus1ix : vkplus1ix + a.sizes.expact - 1), a.sizes.act);
if a.lp.aind == 2, Qkplus1 = Qkplus1'; end;

% put coefficients in place
% size(a.lp.A)
% size(Qkplus1)
% a.lp.m
a.lp.A(:, 1:a.lp.m) = Qkplus1';

[xopt, vopt, flag] = linprog(a.lp.f, a.lp.A, a.lp.b, a.lp.Aeq, a.lp.beq, a.lp.lb, a.lp.ub, [], a.lp.options);
piopt = xopt(1:a.lp.m);
Vkplus1 = min(piopt' * Qkplus1);

qix = ndi2lin([action' a.state'], a.sizes.q);
delta = reward(a.indices(7)) + a.learnparam.gamma * Vkplus1 - a.Q(qix);
a.E(qix) = 1;
a.Q = a.Q + a.learnparam.alpha * delta * a.E;
a.E = a.learnparam.gamma * a.learnparam.lambda * a.E * ~a.exploring;

% we computed an optimal policy in s', we might as well use it
piopt(find(abs(piopt) < 1e-5)) = 0;
piix = ndi2lin([1 state'], a.sizes.pi);
a.PI(piix : piix + a.lp.m - 1) = piopt;


% ==========================================
% == compute and update the policy in current state
% == update the state value as given by the policy

vkix = ndi2lin([ones(1, length(action)) a.state'], a.sizes.q);
Qk = reshape(a.Q(vkix : vkix + a.sizes.expact - 1), a.sizes.act);
if a.lp.aind == 2, Qk = Qk'; end;

% put coefficients in place
a.lp.A(:, 1:a.lp.m) = Qk';

[xopt, vopt, flag] = linprog(a.lp.f, a.lp.A, a.lp.b, a.lp.Aeq, a.lp.beq, a.lp.lb, a.lp.ub, [], a.lp.options);
piopt = xopt(1:a.lp.m);
piopt(find(abs(piopt) < 1e-5)) = 0;
piix = ndi2lin([1 a.state'], a.sizes.pi);
a.PI(piix : piix + a.lp.m - 1) = piopt;

% returning a
% END minimaxq() ============================================================
