function a = plainq(a, state, actions, rewards, params)
%Implements the plain Q-learning algorithm
%  A = PLAINQ(A, STATE, ACTIONS, REWARDS, PARAMS)
%  Implements Q-learning as described in [1], employing in
%  addition to that an eligibility trace. Expects the Q table to be
%  initialized. Uses flat Q and eligibility tables for fast access.
%
%  Required values on the agent learning parameters:
%   alpha           - the learning rate
%   gamma           - the discount factor
%   lambda          - the eligibility trace decay rate
%   epsilon         - the exploration probability
%  Required values on the extra parameters:
%   newtrial        - only in episodic environments; whether a new trial is
%                   beginning
%
% Can be coupled with an action function that uses a Q-table indexed on
% agent state and action, such as greedyact().
%
%  Supports discrete states and actions, with 1 action variable per agent.
%
%
%  References:
%  [1] Watkins, C. J. C. H. and Dayan, P. (1992). Technical note:
%      Q-learning. Machine Learning Journal, 8:279-292.
%
%  See also learn, plainq_init

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


if a.learnparam.episodic && params.newtrial,
    a.E = 0 * a.Q;          % clear eligibility trace
end;

ix = ndi2lin([1 state(a.indices(1) : a.indices(2))'], a.qsize);
qix = ndi2lin([actions(a.indices(5)) a.state(a.indices(1) : a.indices(2))'], a.qsize);

delta = rewards(a.indices(7)) + a.learnparam.gamma * max(a.Q(ix : ix + a.actionscount - 1)) - a.Q(qix);
a.E(qix) = 1;
a.Q = a.Q + a.learnparam.alpha * delta * a.E;
a.E = a.learnparam.gamma * a.learnparam.lambda * a.E * ~a.exploring;

% returning a
% END plainq() ============================================================
