function a = wolfq_init(a, info)
%Initialization of the WoLF-PHC learning algorithm
%  A = WOLFQ_INIT(A, INFO)
%  Initializes the Win-or-Learn-Fast Policy Hill Climbing algorithm.
%
%  Required values on the info parameters:
%   statespacesize      - the size of the state space
%
%  Values on the learning parameters:
%   usefullstate        - whether to learn on the basis of the full state.
%           Default 1 (use full state)
%
%  See also agent_initlearn, wolfq
%
%  Author:      Lucian Busoniu
%  Version:     1.1
%  History:     1.1 added full state

% Check states and actions
if any(info.statespacesize(:, 1) ~= 1) || ...
        a.indices(5) < a.indices(6) || ...
        info.actionspacesize(a.indices(5), 1) ~= 1
    error('wolfq() only supports discrete states and actions, with 1 action variable');
else
    a.actionscount = info.actionspacesize(info.imap(5), 2);
end;

DEF_LP.usefullstate = 1;
DEF_LP.alphadecay = 1;
DEF_LP.deltadecay = 1;
a.learnparam = checkparams(a.learnparam, DEF_LP);

if a.learnparam.usefullstate, a.xindices = 1:size(info.statespacesize, 1);
else a.xindices = a.indices(1):a.indices(2);
end;

% construct the sizes structure; (helper) the state space size...
ss = info.statespacesize(a.xindices, 2)';
% ...the Q-table size, actions vary fastest...
a.sizes.q = [a.actionscount ss];
% ...the size of the policy table, action varies fastest...
a.sizes.pi = a.sizes.q;
% ...the size of the state visits counts
a.sizes.c = ss;

% === Algorithm data for the agent ===

% Qi(s, a1, ..., an) = 0 for all s, a1, ..., an; same for Ei
% Ci(s) = 0 for all s
a.Q = zeros(prod(a.sizes.q), 1);
a.E = a.Q;
a.C = zeros(prod(a.sizes.c), 1);

% PIi(s, ai) = 1 / |Ai|, for all ai
% same for PIavg
a.PI = 1/a.actionscount * ones(prod(a.sizes.pi), 1);
a.PIavg = a.PI;

a.volatile = {'Q', 'E', 'C', 'PIavg'};

% returning a
% END wolf_init()
% ========================================================