function [stats, worlds] = processexp(datafile, mode, plotcount, plotfields, cfg)
% processes a series of experiments ran using runexp()
% STATS = PROCESSEXP(DATAFILE[, MODE, PLOTCOUNT, PLOTFIELDS, CFG])
%  Processes a series of experiments ran using the MA-RL toolbox
%  via the function runexp().
%
%  Parameters:
%   DATAFILE    - the name of the datafile where the experiment results
%       were saved.
%   MODE        - the processing mode. One of four available modes, 'plot',
%       'replay', 'replayeach', 'manual'. See below for a detailed
%       description of each. The default is 'plot'.
%   PLOTCOUNT   - in 'plot' mode, how many plots should be displayed on a
%       single figure. Useful when the experiments sequence is composed of
%       constant-length subsequences each characterizing one category of 
%       settings (e.g. one such subsequence could be characterized by a
%       given configuration of the world, and within it various learning
%       parameter configurations might be tried). The default is 0, which
%       means all plots are collected on a single figure. Use 1 to give
%       each plot its separate figure.
%   PLOTFIELDS  - which fields of the saved statistics are interesting for
%       plotting. A cell array of field names. Default is {'iter'}. Field
%       evolutions will be plotted in separate figures, one per field.
%   CFG         - additional configuration fields (see below for commented
%       defaults)
%
%  Returns:
%   STATS       - replay statistics, in replay/replayeach modes.
%       Either an nexp-cell array of stat structures, in replay mode;
%       or an nexp-cell array of nrun-cell arrays of stat structures, in
%       replayeach mode.
%   WORLDS      - world objects, in replay/replayeach modes.
%       An nexp-cell array of world structures
%
%  The function outputs the configuration of each experiment at the console
%  before processing it.
%
%  Processing modes
% ==================
%  In 'plot' mode, statistics of the learning process are being plotted.
%  The values are averaged over the number of saved independent runs of
%  each experiment.
%  In 'replay' mode, if objects have been stored during running the
%  experiments, the learned behaviour of the agents is replayed. A random
%  or specified experiment run is chosen from the saved independent runs.
%  In 'replayeach' mode, if objects have been stored, the learned behaviour
%  of the agents in each run of each experiment is replayed.
%  In 'manual' mode, the user is given full control of the processing
%  sequence. A command prompt is displayed, where for each experiment the
%  following options are available:
%   p   - [Plot] the experiment statistics in the current figures, open 
%       new figures if none are open.
%   np  - [New figure(s) & Plot] open new figure(s) for the interesting
%       fields, then plot the experiment statistics.
%   r   - [Replay] behaviour from a random run.
%   re  - [Replay Each] replay behaviour from each run of the experiment.
%   pa  - [Plot All] terminates manual mode and processes all the remaining
%       experiments in 'plot' mode.
%   ra  - [Replay All] terminates manual mode and processes all the
%       remaining experiments in 'replay' mode.
%   q   - [Quit] terminates processing discarding any remaining experiments.
%
%  For the 'replay' and 'replayeach' modes, the world must support a
%  view.
%
%
% See also runexp

%  Author:      Lucian Busoniu
%  Version:     1.1
%  History:     15 feb 2006, updated to conform to new style of creating
%               agents


DEF_MODE = 'plot';
DEF_PLOTCOUNT = 0;
DEF_PLOTFIELDS = {'iter'};
REPLAY_SPEED = 7;
INITIAL_MANUAL_OPTION = 'p';

% default additional config
CFG = struct;
CFG.replayrun = [];         % to force a specific run to be replayed, instead of a random one
CFG = setfigprop(CFG, 'addfields');
CFG.styleindex = [];        % can be used in combinaton with addtocrt; [] = default style index
CFG.plotconf = 1;
CFG.plotminmax = 0;
CFG.addtocrt = 0;
CFG.confsty = 'area';       % only area supported currently

% plot styles
PSTYLES = {{'k'}, {'--k'}, {'-.k'}, {':k', 'LineWidth', 2}, ...
    {'r'}, {'--r'}, {'-.r'}, {':r', 'LineWidth', 2}, ...
    {'b'}, {'--b'}, {'-.b'}, {':b', 'LineWidth', 2}};
PATCHSTY = {[.7 .7 .7], 'edgecolor' , 'none', 'facealpha', 1};
LABELS.iter = 'Number of steps';
LABELS.trials = 'Trial number';


% options and their descriptions
mopt = {'p', 'np', 'r', 're', 'pa', 'ra', 'q'};
moptdesc = {'Plot', 'New fig. & Plot', 'Replay', 'Replay Each', 'Plot All', 'Replay All', 'Quit'};
% form the user prompt for manual mode
prompt = sprintf('\nruns');
for i=1:length(mopt),
    prompt = [prompt mopt{i} ' [' moptdesc{i} ']'];
    if i < length(mopt), prompt = [prompt ', ']; end;
end;

if nargin < 2, mode = DEF_MODE; end;
if nargin < 3, plotcount = DEF_PLOTCOUNT; end;
if nargin < 4, plotfields = DEF_PLOTFIELDS; end;
if nargin < 5, cfg = CFG; else cfg = parseconfig(cfg, CFG); end;


% =======================
% == Load data
load(datafile, 'experiments');

% some helper variables
nexp = length(experiments);       % experiments count
nfld = length(plotfields);        % plot fields count
% plot-related
figh = cell(nfld, 1);     % handles to current figures
plotix = 0;                     % current plot index

% init global-experiment vars
switch mode,
    case {'replay', 'replayeach'},
        stats = cell(nexp, 1);
        world = cell(nexp, 1);
end;

% =======================
% == Experiments loop
iexp = 1; lastopt = INITIAL_MANUAL_OPTION;
while iexp <= nexp,
    
    % retrieve experiment data
    exp = experiments{iexp};

    disp(' ');  %  
    disp(['Processing experiment #' num2str(iexp) ' [' exp.config.options.label '] in ' mode ' mode.']);
    disp(['World type: ' exp.config.worldtype]);
    disp('World parameters: ');
    disp(exp.config.worldargs);
    disp('Learning configuration: ');
    disp(exp.config.lp);
    disp(' ');
    disp(['Agent learning parameters: (' num2str(exp.config.agn) ' agents): ']);
    disp(exp.config.aglp);
    disp(' ');
    disp(['Agents functions (' num2str(exp.config.agn) ' agents): ']);
    disp(exp.config.agfun);
    
    nruns =  exp.config.options.nruns;
    
    % reset flags
    manual = 0; newfig = 0; plotting = 0; replaying = 0; repeach = 0; runtime = 0;
    
    % process modes and set flags/init vars accordingly
    switch mode,
        case 'manual',
            manual = 1;
            % get user option (empty defaults to last chosen)
            opt = input([prompt ' <' lastopt '> $ '], 's');
            if isempty(opt), opt = lastopt; end;
            % process user option
            switch opt,
                case 'p', plotting = 1;
                case 'np', plotting = 1; newfig = 1;
                case 'irun', replaying = 1;
                case 're', replaying = 1; repeach = 1;
                case 'pa', plotting = 1; mode = 'plot';
                case 'ra', replaying  = 1; mode = 'replay';
                case 'q', return;
                otherwise, disp('! Option is invalid !'); continue;
            end;
            lastopt = opt;
            
        case 'plot',
            plotting = 1;
            
        case 'runtime',
            runtime = 1;
            
        case 'replay',
            replaying = 1;
            stats{iexp} = [];
            worlds{iexp} = [];
            
        case 'replayeach',
            replaying = 1;
            repeach = 1;
            stats{iexp} = cell(nruns, 1);
            worlds{iexp} = [];
            
    end;

    % =====================
    % == Handle plotting
    if plotting,
        
        % if adding to current, get the figure handle
        if cfg.addtocrt,
            if nfld > 1, error('Cannot add to current figure when there are multiple fields to plot');
            else figh{1} = gcf;
            end;
        else
            % else:
            % we need to spawn new figures if either we haven't got any yet, or
            % it has been enforced by the user above, or it's being dictated by
            % a nonzero plotcount
            newfig = (plotix == 0) ...
                || (manual && newfig) ...
                || (~manual && (plotcount > 0) && (plotix > plotcount));
            % spawn new figures if necessary
            if newfig,
                for ifld = 1:nfld, figh{ifld} = figure(); end;
                plotix = 1;
            end;
        end;
        
        % plot the stuff
        % collect values in an accummulator
        values = cell(nfld, 1);
        lenmin = Inf + zeros(nfld, 1);
        for ifld = 1:nfld,
            for irun = 1:nruns,
                lencrt = length(exp.stats{irun}.(plotfields{ifld}));
                lenmin(ifld) = min(lenmin(ifld), lencrt);
                values{ifld}(irun, 1:lencrt) = exp.stats{irun}.(plotfields{ifld});
            end;
        end;
        % update plots & legends
        for ifld = 1:nfld,
            % pick up the values for this field (up to the maximum length)
            v = values{ifld}(:, 1:lenmin(ifld));
            vmean = mean(v, 1);
            vconf = 1.96 * std(v, 1) ./ sqrt(nruns);
            xval = 1:lenmin(ifld);
            % create/make corresponding figure current
            figure(figh{ifld});    
            % get any existing legend
            [legh, objh, lh, leg] = legend; clear legh objh;
            % plot with corresp style, adding to legend
            if cfg.addtocrt && ~isempty(cfg.styleindex),
                style = PSTYLES{cfg.styleindex + 1};
            else
                style = PSTYLES{mod(plotix - 1, length(PSTYLES)) + 1};
            end;
            lh(end+1) = feval(cfg.plotfun, vmean, style{:});
            if nruns > 1,
                leg{end+1} = [exp.config.options.label ', mean #steps'];
            else
                leg{end+1} = [exp.config.options.label ', #steps'];
            end;
            hold on;
            if nruns > 1 && cfg.plotconf,
                lh(end+1) = patch([xval fliplr(xval)], [vmean-vconf fliplr(vmean+vconf)], -.1 * [xval xval], PATCHSTY{:});
                leg{end+1} = '95% confidence interval';
            end;
            if nruns > 1 && cfg.plotminmax,
                lh(end+1) = feval(cfg.plotfun, max(v), style{:});
                leg{end+1} = [exp.config.options.label, ', max'];
                lh(end+1) = feval(cfg.plotfun, min(v), style{:});
                leg{end+1} = [exp.config.options.label, ', min'];
            end;
            % and set legend
            legend(lh, leg);
            xlabel(LABELS.trials);
            ylabel(LABELS.(plotfields{ifld}));
            % note this assumes only 1 quantity is being plotted 
            % (otherwise the last quantity will overwrite the files)
            setfigprop(cfg);
            saveplot(gcf, [cfg.savedir cfg.savefig], cfg.plottarget);
        end;
        
        plotix = plotix + 1;
    end;

    
    % =====================
    % == Handle runtime
    if runtime,
        % collect values in an accummulator
        rt = zeros(nruns, 1);
        for irun = 1:nruns, rt(irun) = exp.stats{irun}.elapsed_time / sum(exp.stats{irun}.iter); end;
        % display
        disp(sprintf('runtime/step = %.5f; 95%% confint [%.5f, %.5f]', ...
            mean(rt), mean(rt) - 1.96*std(rt)/sqrt(nruns), mean(rt) + 1.96*std(rt)/sqrt(nruns)));
    end;
    
    % =====================
    % == Handle replaying
    if replaying,
        if ~strcmp(exp.config.options.storeobjects, 'on'),
            disp('! Objects not stored for this experiment. Cannot replay. !');
            iexp = iexp + 1;
            continue;
        end;
        
        % recreate world using a template agents set
        for i = 1:exp.config.agn,
            if iscell(exp.config.aglp), aglp = exp.config.aglp{i};        % lps per agent
            else aglp = exp.config.aglp;
            end;
            if iscell(exp.config.agfun{1}), agfun = exp.config.agfun{i};  % funs per agent
            else agfun = exp.config.agfun;
            end;
            agents{i} = agent(i, aglp, agfun{:});
        end;
        exp.config.worldargs{1} = agents;
        world = feval(exp.config.worldtype, exp.config.worldargs{:});
        world = feval([exp.config.worldtype '_display'], world, 1);
        
        if repeach, % replay each experiment run
            for i = 1:nruns,
                [ww, aa, ss] = replay(world, exp.agents{i}, REPLAY_SPEED, exp.config.lp.maxiter);
                stats{iexp}{i} = ss;
            end;
            worlds{iexp} = ww;
        else        % replay an experiment run at random
            if isempty(cfg.replayrun),
                i = ceil(rand * nruns);
            else
                i = cfg.replayrun;
            end;
            [ww, aa, ss] = replay(world, exp.agents{i}, REPLAY_SPEED, exp.config.lp.maxiter);
            stats{iexp} = ss;
            worlds{iexp} = ww;
        end;
        
        % discard the world object
        % disabled since it may be used externally; instead, just hide it
        % feval([exp.config.worldtype '_destroy'], world);
        feval([exp.config.worldtype '_display'], world, 0);
    end;
    
    % next experiment
    iexp = iexp + 1;
end;

