function runexp(expconfigs, datafile, datadir, nruns, silent)
% Runs a sequence of experiments using the MARL framework
% RUNEXP(EXPCONFIGS[, DATAFILE, DATADIR, NRUNS, SILENT])
%  Runs a sequence of experiments using the reinforcement learning framework.
%  Parameters:
%    EXPCONFIGS     - the configuration of the experiments (see below).
%    DATAFILE       - (optional) the name of the data file where to store
%        the results. Default: an automatically generated unique filename.
%    DATADIR        - (optional) the directory where the data file should
%       be saved, if different from the current directory. Set to empty
%        string, has the same effect as not being specified. Default: the
%        current directory.
%    NRUNS          - (optional) the number of independent learning runs
%       for each individual experiment. This number can also be set on a
%       per-experiment basis via the experiment options. If set to anything
%       less than 1, the default will be enforced. Default: 50
%    SILENT         - (optional) 'on'/'off' - silences all text and
%       graphical output when 'on'. Silence can also be set on a
%       per-experiment basis via the  experiment options. However, this
%       global value, if 'on', overrides the per-experiment setting.
%       Default: 'off'.
%
%  EXPCONFIGS is a cell array, every element specifying the configuration
%  of an individual experiment. It is not necessary to specify all the
%  configuration fields for each experiment; anything that is not specified
%  for a given experiment is inherited from the previous one (except the
%  options). There are initial defaults, but it is recommended to specify a
%  full configuration for the first experiment, to maintain consistency
%  among the experiments.
%
%  An experiment configuration specifier is a structure with the following
%  optional fields:
%    worldtype      - the type of the world where the experiment will run.
%    worldargs      - the arguments for the construction of the world. The
%       first argument will be overwritten by the agents collection. A
%       simple empty matrix may be supplied on its position.
%    lp             - the learning parameters. See learn()
%    agn            - the number of agents
%    aglp           - the agent learning parameters. See agent()
%    agfun          - the agent functions, a cell array of strings, in the
%       order: learning function, action function, exploration function.
%       See agent()
%    options        - options for running the experiment.
%
%  The 'alp' and 'agentfuns' can either be structures, in which case all
%  agents will be constructed with the same learning parameters and
%  functions, or can be cell arrays with the same length as the number of
%  agents, in which case each agent gets its own learning parameters and
%  functions.
%
%  The following options are available:
%    nruns      - override the default number of independent runs for
%       the experiment. Default is either 50 or the value specified on the
%       function parameters.
%    silent     - 'on'/'off' silence the experiment. No text or graphical
%       output will be displayed over the duration of the experiment.
%       Default is off.
%    show       - 'on'/'off' whether to display the world view
%       while learning. Default is off. If silent, this option has no
%       effect. Obviously, the world must have a graphical representation
%       if this option is used.
%    storeobjects - 'on'/'off' whether to store the objects after each
%       learning run. Worlds and cleaned up agents will get stored if this
%       option is on. Default is on.
%    convplot   - whether to display convergence summaries after each
%       experiment run. Default is off. If silent, this option has no
%       effect.
%    plotpause  - how long to keep the plots on the screen. Default is
%       three seconds.
%
% See also processexp

%  Version:     1.1
%  History:     15 feb 2006, updated to conform to new style of creating
%               agents


DEF_NRUNS = 50;
DEF_DATAFILE = ['expdata-' num2str(ceil(rand * 1000))];

if nargin <= 1, datafile = DEF_DATAFILE; end;
if nargin <= 2, datadir = ''; end;
if nargin <= 3, nruns = DEF_NRUNS; 
elseif nruns < 1, nruns = DEF_NRUNS;
end;
if nargin <= 4, silentoverride = 'off';
else silentoverride = silent;
end;

% field initial values
INITIAL_CFG = struct(...
    'worldtype', 'gridworld', ...
    'worldargs', {{[], 'Default World', [5 5]', [5 5; 1 5]'}}, ...
    'lp', struct('trials', 250, 'maxiter', 300, 'convgoal', -1, 'avgtrials', 30), ...
    'agn', 2, ...
    'aglp', struct('alpha', .3, 'gamma', .95, 'lambda', .5, 'epsilon', .3), ...
    'agfun', {{'plainq', 'greedyact', 'randomexplore'}}...
);
FIELD_NAMES = fieldnames(INITIAL_CFG);


% option defaults
DEF_OPTIONS = struct(...
    'nruns', nruns, ...
    'silent', 'off', ...
    'show', 'off', ...
    'storeobjects', 'on', ...
    'convplot', 'off', ...
    'plotpause', 3, ...
    'label', '');
OPTION_NAMES = fieldnames(DEF_OPTIONS);

% change to data dir if required
if ~isempty(datadir), crtdir = cd; cd(datadir); end;

experiments = cell(length(expconfigs), 1);

cfg = INITIAL_CFG;
for eix = 1 : length(expconfigs),
    e = expconfigs{eix};
    % process experiment fields, recording changes
    changes = struct();
    for i = 1:length(FIELD_NAMES),
        if isfield(e, FIELD_NAMES{i}),
            cfg.(FIELD_NAMES{i}) = e.(FIELD_NAMES{i});
            changes.(FIELD_NAMES{i}) = '[*]';
        else changes.(FIELD_NAMES{i}) = '';
        end;
    end;
    % process options
    options = DEF_OPTIONS;
    if isfield(e, 'options'),
        for i = 1:length(OPTION_NAMES),
            if isfield(e.options, OPTION_NAMES{i}),
                options.(OPTION_NAMES{i}) = getfield(e.options, OPTION_NAMES{i});
            end;
        end;
    end;
    % create default label if necessary and save options
    if isempty(options.label), options.label = ['Experiment ' num2str(eix)]; end;
    cfg.options = options;
    
    % dump configuration if not silenced
    if ~strcmp(silentoverride, 'on') && ~strcmp(options.silent, 'on'),
        disp(' ');
        disp(['Running experiment #' num2str(eix) ' with configuration: ']);
        disp('Options:');
        disp(options);
        disp(' ');
        disp(['World type: ' cfg.worldtype ' ' changes.worldtype]);
        disp(['World parameters: ' changes.worldargs]);
        disp(cfg.worldargs);
        disp(['Learning configuration: ' changes.lp]);
        disp(cfg.lp);
        disp(' ');
        disp(['Agent learning parameters(' num2str(length(cfg.agn)) ' agents): ' changes.aglp]);
        disp(cfg.aglp);
        disp(' ');
        disp(['Agent functions (' num2str(length(cfg.agn)) ' agents): ' changes.agfun]);
        disp(cfg.agfun);
    end;
    
    % init the experiment structure
    experiments{eix} = struct();
    experiments{eix}.config = cfg;
    experiments{eix}.stats = cell(1, options.nruns);
    if strcmp(options.storeobjects, 'on'),
        experiments{eix}.agents = cell(1, options.nruns);
    end;
    
    % create the template agents set
    agents = cell(1, cfg.agn);
    for i = 1:cfg.agn,
        if iscell(cfg.aglp), aglp = cfg.aglp{i};        % lps per agent
        else aglp = cfg.aglp;
        end;
        if iscell(cfg.agfun{1}), agfun = cfg.agfun{i};  % funs per agent
        else agfun = cfg.agfun;
        end;
        agents{i} = agent(i, aglp, agfun{:});
    end;

    % store templates of agents onto world creation parameters
    cfg.worldargs{1} = agents;
    
	disp('Currently executing run #000/000');
    % run the experiment nruns times
    for r = 1:options.nruns,
        
        if ~strcmp(silentoverride, 'on') && ~strcmp(options.silent, 'on'),
            disp([8*ones(1, 8) sprintf('%4d/%4d', r, options.nruns)]);
        end;
        
        % create the world with the agents in it
        world = feval(cfg.worldtype, cfg.worldargs{:});
        % store the world (once only) if objects storage is on
        if r == 1 && strcmp(options.storeobjects, 'on'),
            experiments{eix}.world = world;
        end;
        % show the world if display is on
        if strcmp(options.show, 'on')...
           && ~strcmp(silentoverride, 'on') && ~strcmp(options.silent, 'on'),
            world = feval([cfg.worldtype '_display'], world, 1);
        end;
        
        % run learning
        [world, learnedagents, stats] = learn(world, agents, cfg.lp);
    
        % record outcomes
        experiments{eix}.stats{r} = stats;
        if strcmp(options.storeobjects, 'on'),
            for i = 1:cfg.agn,
                learnedagents{i} = agent_cleanup(learnedagents{i});
            end;
            experiments{eix}.agents{r} = learnedagents;
        end;
        
        % save the intermediary data
        save(datafile, 'experiments');
        
        % destroy world
        feval([cfg.worldtype '_destroy'], world);
        
        % plot convergence data if required
        if strcmp(options.convplot, 'on') ...
           && ~strcmp(silentoverride, 'on') && ~strcmp(options.silent, 'on'),
            if r == 1,
                avgiter = stats.iter;
            else,
                l = min([length(avgiter), length(stats.iter)]);
                avgiter = (avgiter(1:l) * (r - 1) + stats.iter(1:l)) / r;
            end;
            figure(99); set(99, 'Name', ['Experiment Convergence [' options.label ']'], 'NumberTitle', 'off');
            subplot(211); plot(avgiter, 'k');
            title('Average Convergence over Experiment'); xlabel('Trials'); ylabel('Steps'); grid;
            subplot(212); stairs(stats.iter);
            title('Convergence on Last Run'); xlabel('Trials'); ylabel('Steps'); grid;
            pause(options.plotpause);
            close(99);
        end;
                
    end;    % runs iteration
    
    % garbage collect, it can get quite messy during an experiment
    java.lang.System.gc();
    
end;        % experiments iteration
    

if ~isempty(datadir), cd(crtdir); end;

% END runexp() ===========================================================