function [world, agents, stats] = episodic_replay(world, agents, speed, maxiter)
%Replays the agents' current policies in an episodic environment
%  [WORLD, AGENTS, STATS] = EPISODIC_REPLAY(WORLD, AGENTS, SPEED, MAXITER)
%  Parameters:
%   WORLD       - the world where the agents live
%   AGENTS      - the (possibly heterogeneous) cell array of agents
%   SPEED       - the speed of the replay, between 1 and 10
%   MAXITER     - for how many iterations to allow the policy to run at
%                 most. May be -1 ('run forever').
%  Returns:
%   WORLD       - the possibly altered world.
%   AGENTS      - the possibly altered agents.
%
%  Replays the agents' policies over a single trial. Assumes that the
%  agents have already been correctly setup during the learning process.
%  Requires that the world implements a view. This view will be shown if
%  not already shown.
%
%
%  See also replay, episodic_learn

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


% after how many iterations during which the state does not change the
% agents are considered stuck and the replay is interrupted
STUCK_ITER = 12;     

% shorthand constants
wresetfun = strcat(world.type, '_reset');
wdynamicsfun = strcat(world.type, '_dynamics');
vdisplayfun = strcat(world.type, '_display');
winfofun = strcat(world.type, '_info');
if speed == 0, pauselen = -1;
else pauselen = 1 - min(max(speed, 1), 10) / 10;
end;

% retrieve world info and initialize needed data structures
info = feval(winfofun, world);
nagents = length(agents);
aspan = 1:nagents;
nactions = size(info.actionspacesize, 1);           % how many action variables in total
actspans = cell(nagents, 1);                        % positions on the joint action
for i = 1:nagents,
    actspans{i} = info.imaps(5, i) : info.imaps(6, i);
end;
p = length(info.statespacesize);
q = length(info.actionspacesize);

% make sure world is displayed
world = feval(vdisplayfun, world, 1);

% initialize loop
action = zeros(nactions, 1);                    % joint action
actionviews = nan(nactions, nagents);           % undefined last action views
rewardviews = nan(nagents, nagents);            % undefined last reward views
finished = zeros(1, nagents);                   % initially noone finished

[world, stateviews] = feval(wresetfun, world, 0);
if pauselen < 0, pause;
else pause(pauselen); end;

% init history
if maxiter == -1, Ns = 100; else Ns = maxiter + 1; end;
t = 0 : (Ns - 1);
x = zeros(p, Ns); x(:, 1) = stateviews(:, i);  % state assumed == 1st agent view
u = zeros(q, Ns);
r = zeros(nagents, Ns); r(1) = NaN;

stuckcheck = [];
params.donotlearn = 1;
params.newtrial = 1; k = 1; stuck = 0;

while maxiter == -1 || k <= maxiter,
    % controllers evolve first
    for i = aspan,
        params.finished = finished(i);
        [agents{i} action(actspans{i})] = agent_control(agents{i}, ...
            stateviews(:, i), actionviews(:, i), rewardviews(:, i), params);
    end;
    % terminate if all finished...
    if finished, break; end;
    % ...or the agents are stuck (does not work for loops)
    % WARNING! works only in environments where all 
    % the agents' state views are identical for identical states
    if size(stuckcheck, 2) == STUCK_ITER,
        if ~any(diff(stuckcheck, 1, 2)), stuck = 1; break; end;
        stuckcheck = [stuckcheck(:, 2:size(stuckcheck, 2)) stateviews(:)];
    else
        stuckcheck = [stuckcheck stateviews(:)];
    end;
    % plant evolves
    tt = cputime;
    [world, stateviews, actionviews, rewardviews, finished] = feval(wdynamicsfun, world, action);
    tt = cputime - tt;
    % pause taking also into account the time spent inside the display
    if pauselen < 0, pause;
    elseif pauselen > tt, pause(pauselen - tt); end;
    % add to history    
    u(:, k) = action; x(:, k+1) = stateviews(:, 1); r(:, k+1) = rewardviews(:, 1);
    params.newtrial = 0; k = k + 1;
end;

% form statistics, finalizing and adding history
k = k - 1; Ns = k; 
stats.iter = Ns; stats.Ns = Ns;      % alias 
stats.stuck = stuck;
u(:, k) = NaN;      % last action is undefined
stats.t = t(1:Ns+1); stats.x = x(:, 1:Ns+1); stats.u = u(:, 1:Ns+1); stats.r = r(:, 1:Ns+1);
clear t x u r 
% Note a standard history would include the return; here it's more
% complicated because we have several agents; for now, don't compute the
% return

% returning world, agents, stats
% END replay() ============================================================