function replay_algorithm(task, cooplev, method, cfg)
% REPLAY_ALGORITHM Runs a given MARL algorithm on a given type of task.
%  REPLAY_ALGORITHM(TASK, METHOD, CFG)
%
%  Replays a MARL algorithm which was previously run using run_algorithm
%  on a given type of task. Specifically designed for use with alg_gui, but
%  can also be used independently. Retrieves the algorithm data from the
%  field of the 'algdata' structure in the base workspace created by
%  run_algorithm. A warning message is displayed if the data for the
%  task/algorithm combination is not found.
%
%  Parameters:
%   TASK        - specifies the type of task. All task types supported by
%       run_algorithm are supported here.
%   COOPLEV     - specifies the cooperation level. All levels supported by
%       run_algorithm are supported here.
%   METHOD      - specifies the algorithm to replay. All algorithms
%       supported by run_algorithm are supported here.
%   CFG         - configuration of learning process. Structure with the
%       following fields (all optional)
%       'replayspeed'    - Replay speed, between 0 and 10. Default 6.
%   
% See also run_algorithm

% Author: Lucian Busoniu
% Version: 2.0
% History: 6 Sept 2006, 2.0 -- added support for robot rescue gridworld


DEF_CFG.replayspeed = 6;
if nargin < 3, cfg = DEF_CFG;
else cfg = checkparams(cfg, DEF_CFG); end;

algdata = evalin('base', 'algdata');
% First, make sure all worlds are hidden
fn = fieldnames(algdata);
for i = 1:length(fn),
    try
        wtype = algdata.(fn{i}).world.type;
    catch
        warning('Could not replay; no algorithm was yet run');
        return;
    end;
    algdata.(fn{i}).world = feval([wtype '_display'], algdata.(fn{i}).world, 0);
end;

datvar = [task '_' cooplev '_' method];

if ~isfield(algdata, datvar),
    warning(['Could not replay; [' method '] was not yet run in [' cooplev '] setting on task [' task ']']);
    return;
end;
dat = algdata.(datvar);

disp(['Replaying policies learned with [' method '] at cooperation level [' cooplev '] on task [' task ']']);

% Then replay the correct world & agents
[dat.world dat.agents repstats] = replay(dat.world, dat.agents, cfg.replayspeed);

% update and save data
algdata.(datvar) = dat;
assignin('base', 'algdata', algdata);

if repstats.stuck,
    disp('*** The agents got stuck! Trial interrupted. ***');
else
    disp(['Trial finished in ' num2str(repstats.iter) ' steps.']);
end;


% END replay_algorithm() ----------------------------------------------