function run_algorithm(task, cooplev, method, cfg)
% RUN_ALGORITHM Runs a given MARL algorithm on a given type of task.
%  RUN_ALGORITHM(TASK, COOPLEV, METHOD, CFG)
%
%  Runs a MARL algorithm on a given type of task. Uses 5x5 robots
%  gridworlds as tasks. Specifically designed for use with alg_gui, but can
%  also be used independently. Saves the algorithm data in a field of the
%  'algdata' structure in the base workspace. Creates the variable if not
%  present. The field name is created as a combination of task type and
%  algorihtm type. Any previous data for that task/algorithm pair is
%  cleared and the resources associated with it released.
%
%  Parameters:
%   TASK        - specifies the type of task. One of 'nav', 'rescue',
%       for navigation task, or search & rescue task, respectively.
%   COOPLEV     - specifies the cooperation level in the task. One of 'coop', 'comp',
%       'mixed', for fully cooperative, fully competitive, and mixed tasks,
%       respectively.
%   METHOD      - specifies the algorithm to run. One of 'q' (single-agent
%       Q-learning), 'fsq' (full-state Q-learning), 'asf', (adaptive state
%       focus Q-learning), 'teamq' (team Q-learning), 'wolf',
%       (Win-or-Learn-Fast policy hill climbing), 'opt' (predefined optimal
%       sequences on the given task).
%   CFG         - configuration of learning process. Structure with the
%       following fields (all optional)
%       'stopearly'         - if learning should be stopped upon
%           convergence. Default 1 (yes).
%       'showafter'         - make world visible only after this number of
%           trials. If Inf, world will be invisible throughout the learnign
%           process. Default Inf.
%       'slowdownafter'     - similarly, but for slowing down the learning
%           process for visualization. Takes effect only after 'showafter'
%           trials. Default Inf.
%       'pausebeforerun'    - if world should be held on screen until
%           keypress before starting the learning. Default 1 (hold on
%           screen).
%       'useseeds'          - whether the rand generator should be
%           deterministically initialized for reproducible behavior of
%           algorithms. Default 1 (use reproducible behavior).
%   
% See also replay_algorithm

% Author: Lucian Busoniu
% Version: 2.0
% History: 6 Sept 2006, 2.0 -- added support for robot rescue gridworld, removed competitive /
% minimax code

DEF_CFG.stopearly = 1;
DEF_CFG.pausebeforerun = 1;
DEF_CFG.useseeds = 1;

DEF_VIEWCFG.repres = 'abstract';     % no longer present in GUI, always abstract repres
DEF_VIEWCFG.showafter = Inf;
DEF_VIEWCFG.slowdownafter = Inf;
DEF_CFG.viewcfg = DEF_VIEWCFG;

if nargin < 3, cfg = DEF_CFG;
else cfg = checkparams(cfg, DEF_CFG); end;

% task
% cooplev
% method
% cfg
% viewcfg = cfg.viewcfg
% return

% top-level def learning params
Nag = 2;        % two agents - most of the times
lp.trials = 100;
lp.maxiter = 1000;
if cfg.stopearly, lp.convgoal = 0.001;
else lp.convgoal = -1;
end;
lp.avgtrials = 10;
lp.showconv = 1;        % show online conv plot for all algorithms
lp.closeconv = 0;

% top-level default agent learning params
alp.alpha = .1;
alp.gamma = .98;
alp.lambda = 0.5;
alp.epsilon = .7;

% init structs
seeds = struct();
OVWR = struct();


% ---------------------------
% By task type and cooperation level
% setup task structure + reward function
% setup common learning params
% setup learning params overwrites

switch task,
    
    % ------------------------------------------
    % Navigation task
    case 'nav',
        
        % common learning parameters/overwrites for navigation task
        alp.window = 256;   % for ASF
        lp.trials = 100;
        lp.maxiter = 1000;
        alp.alpha = .2;
        alp.gamma = .98;
        alp.lambda = 0.5;
        alp.epsilon = .7;
        
        % A set of optimal action sequences
        OPTSEQ.coop = {'ursruusu', 'ulluurul'};
        OPTSEQ.mixed = {'ursruuurr', 'ulluuull'};

        switch cooplev,
            case 'coop',
                grid =      [1 0 4 0 1;   % Y
                             0 0 0 0 0;   % |
                             1 1 0 1 1;   % |
                             0 0 0 0 0;   % |
                             2 0 0 0 3];   % |
                             % 0 -- -------> X    0
                goal.type = 'coop';
                goal.rew = [-.1 NaN 10];
                % random seeds
                seeds.q         = 1e12;
                seeds.wolf      = 1e6;
                seeds.teamq     = 1e3;
                seeds.fsq       = 1e10;
                seeds.asf       = 1e9;
                % max trials settings, -1 = default
                max_trials.q        = 40;
                max_trials.wolf     = -1;
                max_trials.teamq    = -1;
                max_trials.fsq      = -1;
                max_trials.asf      = -1;
                
            case 'mixed',
                % agent 1: 2, agent 2: 3, goal 1: 4, goal 2: 5
                grid =      [5 0 0 0 4;   % Y
                             0 0 0 0 0;   % |
                             1 1 0 1 1;   % |
                             0 0 0 0 0;   % |
                             2 0 0 0 3];   % |
                             % 0 -- -------> X    0
                goal.type = 'mixed';
                goal.rew = [-.1 -.1 10];
                % random seeds
                seeds.q         = 1e12;
                seeds.wolf      = 1e6;          % 15 steps, but can't get it to work better...
                seeds.teamq     = 1e3;
                seeds.fsq       = 1e10;
                seeds.asf       = 1e9;
                % max trials settings, -1 = default
                max_trials.q        = 40;
                max_trials.wolf     = -1;
                max_trials.fsq      = -1;
                max_trials.asf      = -1;

            otherwise, error(['Unknown cooperation level [' cooplev ']']);

        end;        % cooplev switch for navigation task
        grid = fliplr(grid');

        % process grid matrix into world data structures
        [i j] = find(grid == 1);
        obstacles = [i j]';
        worldsize = size(grid)';
        [x1 y1]   = find(grid == 2); [x2 y2]   = find(grid == 3);
        [xg1 yg1] = find(grid == 4); [xg2 yg2] = find(grid == 5);
        positions = [x1 y1; x2 y2]'; goal.pos = [xg1 yg1; xg2 yg2]';

        % -----------------------------------
        % param overwrites for methods
        
        % custom rewards for teamq
        ovwr.name = 'goal.rew';
        ovwr.value = '[0 NaN 10]';
        OVWR.teamq = {ovwr};

        % END navigation task CASE

    % ------------------------------------------
    % Rescue robotics simulated task
    case 'rescue',
        
        % common lps/overwrites for the rescue task
        lp.avgtrials = 20;       
        % common lps for algorithms on this task
        lp.trials = 100;
        lp.maxiter = 1000;
        alp.alpha = .2;
        alp.gamma = .98;
        alp.lambda = 0.5;
        alp.epsilon = .7;
        
        switch cooplev,

            case 'coop',
        %         % Agents can pass narrowway, have to coordinate upon return to homebase
        %         grid =      [0      0       0       0       0       0;
        %                      0      0       0       101.1   101   101.2;   % |
        %                      0      0       0       -1       -1      -1;   % |
        %                      0      0       0       0       0       2;
        %                      1      10      10      10      -1      0];   % |
        %                      % 0 -- -------> X    0
        %         % random seeds
        %         seeds.q         = 1e10;       % OK, stop near target
        %         seeds.wolf      = 0;          % suboptimal, but valid policy
        %         seeds.fsq       = 1e3;        % works ok, very suboptimal policy
        %         seeds.asf       = 1e10;       % reasonable, tho suboptimal
        %         alp.window = 2048;       % for ASF
                % TODO optimal sequences not defined for this world

                % Homebase at top, victim near homebase
                grid =      [10      10     10      -1      -1;   % Y
                             0       0      101.1    101    101.2;
                             0       0      0       0       0
                             -1     -1      0       -1      -1;   % |
                             0      0       0       0       0;
                             1      0       0       0      2];   % |
                OPTSEQ.coop = {'ursruuusllunnnn', 'ulluurrullunnn'};    % 11 steps
                
                % random seeds
                seeds.q         = 1e3;      % good, wobbles around optimal
                seeds.wolf      = 0;        % works, but suboptimal (+4 steps)
                seeds.teamq     = 1e3;        % does not work properly, too long to solve task +7
%                 seeds.fsq       = 0;        % shows bad conv at start, 4-steps subopt (15 steps); this was tested for alpha=0.2
                seeds.fsq       = 1e24;        % good. shows bad conv at start, 5-steps subopt (16 steps)
                seeds.asf       = 0;        % PERFECT! (i.e. optimal)
                % max trials settings, -1 = default
                max_trials.q        = 40;
                max_trials.wolf     = -1;
                max_trials.teamq    = -1;
                max_trials.fsq      = -1;
                max_trials.asf      = -1;

                wcfg.tasktype = 'coop';
                % Rewards: [step coll lock sit pull rip home]
                wcfg.rew.step   = -1;
                wcfg.rew.coll   = -1;
                wcfg.rew.lock   = 10; 
                wcfg.rew.sit    = -1;
                wcfg.rew.pull   = -1;
                wcfg.rew.rip    = -1;
                wcfg.rew.home   = 100;
                
                % param overwrites for methods
                ovwr.name = 'wcfg.rew';
                % Custom Rewards: [step coll lock sit pull rip home]
                ovwr.value = '[0 0 1 0 0 0 100]';
                OVWR.teamq = {ovwr};
                ovwr.name = 'alp.alpha';
                ovwr.value = '0.05';
                OVWR.teamq{end+1} = ovwr;
                
                ovwr.name = 'alp.alpha';
                ovwr.value = '0.1';
                OVWR.fsq = {ovwr};

                alp.window = 512;           % for ASF

            case 'mixed',
                % Codes:
                % -1 obstacle, 1, 2... agents, 101, 102... targets, 10x.1, 10x.2, ... lock position
                % for target x, 10 homebase area
%                 % Agents can pass narrowway, have to coordinate upon return to homebase
%                 grid =      [-1      0      0      -1        -1        0;   % Y
%                              0      0       0       0       0       0;
%                              101.1  101     0       0        102   102.1;   % |
%                              -1     -1      0       0       -1      -1;   % |
%                              0      0       0       0       0       0;
%                              1      10      10      10      10      2];   % |
%                              % 0 -- -------> X    0
%                 % random seeds
%                 seeds.q         = 1e6;        % OK, proves the point
%                 seeds.wolf      = 0;          % can't get it to work properly...
%                 seeds.fsq       = 1e6;        % too long until convergence...
%                 seeds.asf       = 1e6;        % seems OK
                % TODO optimal sequences not defined for this world

                % Homebase at top, agents have to coordinate at passing thru gate with victim to homebase
                grid =      [0      10     10      10      10       0;  
                             0       0      0       0       0       0;
                             -1     -1      0       0       -1      -1;
                             101.1  101     0       0       102     102.1;
                             0      0       0       -1      0       0;
                             1      0       -1       0      0       2];
                % A set of optimal action sequences
                OPTSEQ.mixed = {'uussrruullunnnn', 'uulluuunnnnnn'}; % 11 steps optimal

                seeds.q         = 1e3;       % OK, proves the point
                seeds.wolf      = 1e6;       % +3 steps
                seeds.fsq       = 1e22;      % OK, learns a little bit too fast (cmp to asf) +4 steps
                seeds.asf       = 0;         % optimal                
                % max trials settings, -1 = default
                max_trials.q        = 40;
                max_trials.wolf     = -1;
                max_trials.fsq      = -1;
                max_trials.asf      = -1;

                wcfg.tasktype = 'mixed';
                % Rewards: [step coll lock sit pull rip home]
                wcfg.rew.step   = -1;
                wcfg.rew.coll   = -1;
                wcfg.rew.lock   = 10; 
                wcfg.rew.sit    = -1;
                wcfg.rew.pull   = -1;
                wcfg.rew.rip    = -1;
                wcfg.rew.home   = 100;

                alp.window = 350;       % for ASF

            otherwise, error(['Unknown cooperation level type' cooplev]);

        end;
        grid = fliplr(grid');

        % process grid matrix into world config
        wcfg.name = 'Robot Search&Rescue';
        [i j] = find(grid == -1);
        wcfg.obstacles = [i j]';
        wcfg.size = size(grid)';

        [i j] = find(grid == 10);
        xll = min(i); yll = min(j); xur = max(i); yur = max(j);
        wcfg.homebase = [xll yll; xur yur]';

        [x1 y1] = find(grid == 1); [x2 y2] = find(grid == 2);
        wcfg.initpos = [x1 y1; x2 y2]';

        targets = {};
        tdelta = 1;
        [xt yt] = find(grid == 100+tdelta);   % goals
        while ~isempty(xt),
            targ.pos = [xt yt]';
            targ.lock = '';
            kdelta = .1;
            [xk yk] = find(grid == 100+tdelta+kdelta);
            while ~isempty(xk),
                dpos = [xk-xt yk-yt];
                if     dpos == [1 0],   lockdir = 'e';
                elseif dpos == [-1 0],  lockdir = 'w';
                elseif dpos == [0 1],   lockdir = 'n';
                elseif dpos == [0 -1],  lockdir = 's';
                else disp('Invalid lock position');
                end;
                targ.lock(end+1) = lockdir;
                kdelta = kdelta + .1;
                [xk yk] = find(grid == 100+tdelta+kdelta);
            end;
            targets{end+1} = targ;
            tdelta = tdelta + 1;
            [xt yt] = find(grid == 100+tdelta);   % goals
        end;
        wcfg.targets = targets;
        
       
        % END rescue task CASE
        
    otherwise,
        error(['Unknown task [' task ']']);

        
end;    % END task type switch


% ----------------------------
% pick method
explorefun = 'randomexplore';       % default; most of the times
runopt = strcmp(method, 'opt');

switch method,
    case 'singleq',
        learnfun = 'plainq';
        actfun = 'greedyact';
        Nag = 1;
    case 'q',
        learnfun = 'plainq';
        actfun = 'greedyact';
    case 'wolf',
        learnfun = 'wolfq';
        actfun = 'pigreedyact';
        alp.alphadecay = [1 500];
        alp.deltadecay = [1 250];
%         alp.deltaw = .2;
        alp.deltaratio = 4;
        alp.usefullstate = 1;
    case 'teamq',
        learnfun = 'teamq';
        actfun = 'teamgreedyact';
        alp.usesocconv = 1;
        alp.lambda = 0;     % to guarantee synchronism of Qtables
        alp.alphadecay = [1 500];   % note this is same as WoLF
        % no neg reward per step, otherwise team Q does not work
    case 'fsq',
        learnfun = 'fullstate_plainq';
        actfun = 'fullstate_greedyact';
    case 'asf',
        learnfun = 'asfq';
        actfun = 'asfq_greedyact';
        alp.expresetweight = 0.1;
    case 'opt',     % static agents with optimal sequences
        learnfun = 'nillearn';
        actfun = 'staticact';
        explorefun = 'nilexplore';
        alp1.static.type = 'sequence';
        alp1.static.value = dirs2codes(OPTSEQ.(cooplev){1});
        alp2.static.type = 'sequence';
        alp2.static.value = dirs2codes(OPTSEQ.(cooplev){2});
        alp = {alp1 alp2};
        
    otherwise,
        error(['Unknown method [' method ']']);
        
end;
% apply the rand seed for reproducibility of results
if isfield(seeds, method), rand('seed', seeds.(method)); end;
if isfield(max_trials, method) && max_trials.(method) > 0, lp.trials = max_trials.(method); end;
% place any overwrites
if isfield(OVWR, method), 
    ovwr = OVWR.(method);
    for i = 1:length(ovwr), eval([ovwr{i}.name '=' ovwr{i}.value ';']); end;
end;

% ------------------------------
% Create agents and world
agents = {};
for i = 1:Nag,
    if iscell(alp),         p = alp{i};         else p = alp; end;    
    if iscell(learnfun),    lf = learnfun{i};   else lf = learnfun; end;
    if iscell(actfun),      af = actfun{i};     else af = actfun; end;
    if iscell(explorefun),  ef = explorefun{i}; else ef = explorefun; end;
    agents{i} = agent(i, p, lf, af, ef);
end;

% create world

switch task,
    case 'nav',
        gwtitle = 'Robots'' World';
        world = robotsgw(agents, gwtitle, worldsize, goal, positions, obstacles, [], cfg.viewcfg);
        
    case 'rescue',
        if Nag == 1,        % need to strip down config
            wcfg.targets = wcfg.targets{1};
            wcfg.initpos = wcfg.initpos(:, 1);
        end;
        % place the view config from the function config onto the world config
        wcfg.viewcfg = cfg.viewcfg;
        wcfg.name = 'Robot Search & Rescue';
        world = rrgw(agents, wcfg);
        
end;

% ------------------------------
% Run learning and save experiment data

% load any prior data
try algdata = evalin('base', 'algdata');
catch
    algdata = struct();
end;

% first, make sure all worlds are hidden
fn = fieldnames(algdata);
for i = 1:length(fn),
    algdata.(fn{i}).world = feval([world.type '_display'], algdata.(fn{i}).world, 0);
end;
% Show current world all the time if no show time has been set
if ~isfinite(cfg.viewcfg.showafter) && ~runopt,
    world = feval([world.type '_display'], world, 1);
end;

% Run learning (unless optimal sequence, in which case it makes no
% sense)
if ~runopt,
    disp(['Running learning on [' task '] task, on cooperation level [' cooplev '], using [' method ']...']);
    if (cfg.pausebeforerun) && feval([world.viewtype '_isvisible'], world.view), pause; end;
    [world, agents, stats] = learn(world, agents, lp);
    disp('Learning complete.');
else        % optimal sequence
    % Need to initlearn() the agents ! Run dummy learning 
    lp.trials = 0; [world, agents, stats] = learn(world, agents, lp);
    % for optimal sequences, no learning has been performed, so we need to pause if requested
    if cfg.pausebeforerun,
        world = feval([world.type '_display'], world, 1);
        pause;
    end;
end;


% experiment data field
datvarname = [task '_' cooplev '_' method];
% cleanup if var with this name already stored
if isfield(algdata, datvarname),
    feval([world.type '_destroy'], algdata.(datvarname).world);
    algdata = rmfield(algdata, datvarname);
end;
% store experiment on data
algdata.(datvarname) = varstostruct('world', 'agents', 'stats');
% save data back to the workspace
assignin('base', 'algdata', algdata);

% replay once
replay_algorithm(task, cooplev, method, cfg);

% END run_algorithm -------------------------------------------------------------------------




% Character directions (actions) to their numeric codes
% In: dseq, char array
% out: cseq, codes array
function useq = dirs2codes(dseq)
map = 'lrdusn'; acts = [1 2 3 4 5 NaN];
useq = zeros(size(dseq));
for i = 1:length(dseq),
    useq(i) = acts(find(map == dseq(i), 1));
end;

% END dirs2codes