%% RRGW_DEMO 
% Script to demonstrate the application of several variants of Q-learning on the 
% robotic rescue gridworld problem 
%
% This script can be run from the command line, from top to bottom, 
% or in cell mode (press CTRL+ENTER within any cell to run it).
% 1. Run the initialization cell first
% 2. Run the algorithm cells 
% 3. Run any evaluation cells
% 4. Run the cleanup cell to delete all the files created by the demo

% This script uses the batch-experiment functions runexp and processexp; to see how to run learning
% algorithms directly, see learndemo

%% Common config

ccfg = struct;
% number of agents
ccfg.agn = 2;               % two agents
% global learning params
ccfg.lp.trials = 70;        % run 70 trials
ccfg.lp.maxiter = 1000;     % each consisting of max 1000 steps
ccfg.lp.convgoal = -1;      % run for exactly lp.trials (difference between Q-tables will never be negative)

% world config, see also the function rrgw
wcfg.name = 'Search & Rescue Gridworld';
wcfg.tasktype = 'coop';     % task is fully cooperative
% configure layout of the world
wcfg.size = [7 6]';
wcfg.obstacles = [4 5; 1 2; 2 2; 4 2; 5 2; 6 2; 7 2; 7 1]';
wcfg.homebase = [3 6; 5 6]';    % homebase is a rectangle with these corners
wcfg.targets.pos = [4 4]';      % target position
wcfg.targets.lock = 'we';       % agents should grab target to the west and east, respectively
wcfg.initpos = [1 1; 5 1]';     % agents are initialized here
% reward is always zero except when locking on target, and bringing it home
wcfg.rew.step   = 0;        % regular step
wcfg.rew.coll   = 0;        % collision
wcfg.rew.lock   = 1;        % lock on target
wcfg.rew.sit    = 0;        % stay put
wcfg.rew.pull   = 0;        % pull on target in the same direction
wcfg.rew.rip    = 0;        % pull on target in different directions
wcfg.rew.home   = 10;       % target is home

% place the world config on the main config
ccfg.worldtype = 'rrgw';
ccfg.worldargs = {[], wcfg};% the first argument is the agents, will be created in runexp

% set other options for the learning experiment
ccfg.options.nruns = 1;             % only repeat each experiment once (this is a demo, we'd get bored otherwise)
ccfg.options.show = 'on';           % enable display of world & agents while learning


%% Single-agent Q-learning

qcfg = ccfg;
qcfg.options.label = 'Single-agent Q-learning'; % will be useful if plotting results
% learning, action, and exploration function
qcfg.agfun = {'fullstate_plainq', 'fullstate_greedyact', 'randomexplore'};
% agent learning parameters
qcfg.aglp.gamma = .98;              % discount factor
qcfg.aglp.alpha = .1;               % learning rate
qcfg.aglp.lambda = 0;               % no elig trace
qcfg.aglp.epsilon = .8;             % initial exploration probability
qcfg.aglp.epsilontrialdecay = .9;   % decay rate pe trial

% run the experiment
runexp({qcfg}, 'rrgwdemo_marl_plainq');


%% Team Q-learning

teamqcfg = ccfg;
teamqcfg.options.label = 'Team Q-learning';
% learning, action, and exploration function
teamqcfg.agfun = {'teamq', 'teamgreedyact', 'randomexplore'};
% agent learning parameters -- same as for Q-learning
teamqcfg.aglp.gamma = .98;
teamqcfg.aglp.alpha = .1;
teamqcfg.aglp.lambda = 0;
teamqcfg.aglp.epsilon = .8;
teamqcfg.aglp.epsilontrialdecay = .9;
teamqcfg.aglp.usesocconv = 0;       % no social conventions

% run experiment
runexp({teamqcfg}, 'rrgwdemo_marl_teamq');


%% Team Q-learning, this time with social conventions
% Note requires the cell above to have been run, since it inherits its config

socconvcfg = teamqcfg;
socconvcfg.options.label = 'Team Q-learning with social conventions';
socconvcfg.aglp.usesocconv = 1;

runexp({socconvcfg}, 'rrgwdemo_marl_teamqsocconv');


%% WoLF-Q-learning

wolfq = ccfg;
wolfq.options.label = 'WoLF-Q-learning';
% learning, action, and exploration function
wolfq.agfun = {'wolfq', 'stochact', 'randomexplore'};
% agent learning parameters
wolfq.aglp.usefullstate = 1;    % we are using the full state signal
% these are the same as above...
wolfq.aglp.gamma = .98;
wolfq.aglp.alpha = .1;
wolfq.aglp.lambda = 0;
wolfq.aglp.epsilon = .8;
wolfq.aglp.epsilontrialdecay = .9;
% ... these are specific to WoLF-PHC
wolfq.aglp.deltaw = .1;         % policy increment when winning
wolfq.aglp.deltaratio = 4;      % ratio of increment when winning / when losing

runexp({wolfq}, 'rrgwdemo_marl_wolfq');



%% Plot evolution of number of steps per trial, for each algorithm
cfg.styleindex = 1;
processexp('rrgwdemo_marl_plainq', 'plot', 0, {'iter'}, cfg);
% add the ones below to the same figure
cfg.addtocrt = 1;
cfg.styleindex = 2;
processexp('rrgwdemo_marl_teamq', 'plot', 0, {'iter'}, cfg);
cfg.styleindex = 3;
processexp('rrgwdemo_marl_teamqsocconv', 'plot', 0, {'iter'}, cfg);
cfg.styleindex = 4;
processexp('rrgwdemo_marl_wolfq', 'plot', 0, {'iter'}, cfg);



%% Replay solution for a couple of algorithms
processexp('rrgwdemo_marl_plainq', 'replay');
processexp('rrgwdemo_marl_wolfq', 'replay');



%% Cleanup

delete rrgwdemo_marl_plainq.mat
delete rrgwdemo_marl_teamq.mat
delete rrgwdemo_marl_teamqsocconv.mat
delete rrgwdemo_marl_wolfq.mat
