function gw = robotsgw(agents, name, worldsize, goal, initpos, obstacles, rewards, cfg)
%Creates a 2D grid world
%  GW = robotsgw(AGENTS, NAME, WORLDSIZE, GOAL, [INITPOS, OBSTACLES, REWARDS, CFG])
%  Parameters:
%   AGENTS      - the agents in the robotsgw, as a (possibly
%               heterogeneous) cell array
%   NAME        - the name of the robotsgw
%   WORLDSIZE   - the size of the 2D robotsgw, a vector [W H]
%   GOAL        - the goal positions of the agents. Can be either a
%   structure or a matrix. If a structure, it contains the following
%   fields:
%       type  - 'mixed', 'coop', or 'comp'
%       pos   - goal position. For mixed settings, one agent's goal position on
%               each row. Should specify goal positions for the maximum
%               number of agents in the robotsgw. For cooperative or
%               competitive settings, one goal position only.
%       rew   - (optional) a vector of reward values, with the structure
%               [ORD COLL GOAL]. ORD will be rewarded for ordinary moves,
%               COLL for collisions, GOAL for reaching the goal. For
%               cooperative goals, COLL is ignored. For competitive goals,
%               ORD and COLL are ignored. This field supersedes the REWARDS
%               argument.
%   If GOAL is an array, a mixed setting is assumed, and GOAL should have the
%   structure of GOAL.pos described above.
%   INITPOS     - (optional) the deterministic initial positions of the
%               agents. If an empty matrix, at each trial start agents will
%               be placed in random positions in the world. One agent
%               position on each column. Should specify initial positions for
%               the maximum number of agents in the robotsgw.
%               Default: empty matrix.
%   OBSTACLES   - (optional) the obstacles to place in the robotsgw.
%               May be: a matrix with obstacle 2D coordinates on each column;
%               an empty matrix, in which case no obstacles are placed; or
%               the string 'user' to obtain the obstacles interactively
%               from the user. Default is empty matrix (no obstacles)
%   REWARDS     - (optional) a vector with three elements, same structure as
%               GOAL.rew. Default is [-1 -2 10]
%   CFG         - (optional) other configuration options
%
%  Agents can move one cell in either of the four compass directions,
%  or stay still. Each agent has one action variable available, and its
%  values are coded in the following way:
%   1 - left
%   2 - right
%   3 - up
%   4 - down
%   5 - stay put
%  Note that an agent can function without the fifth action.
%
%  The agents cannot move out of the grid, onto an obstacle, or in mixed
%  worlds onto a goal that does not belong to them. If agents collide, they
%  will be reset to their positions previous to the collision.
%  Each agent finishes when it has reached its corresponding goal. There is
%  no failure situation. In competitive worlds, the agent that reaches the
%  goal first wins, the other loses, and the trial terminates at that
%  moment.
%
%
%  See also robotsgwview, robotsgw_advance, robotsgw_destroy,
%      robotsgw_obstacles, robotsgw_reset

%  Author:      Lucian Busoniu
%  Version:     2.0
%  History:     2.0 added handling of coop and comp tasks

REQ_ARG_COUNT = 4; DEFAULT_REWARDS = [-1 -2 10];
DEF_CFG = struct();

% verification of required arguments
if nargin < REQ_ARG_COUNT, error('Incorrect number of inputs'); end;

% handle variable argument lists, transforming into the row format
if nargin < REQ_ARG_COUNT + 1, gw.initpos = []; 
else gw.initpos = initpos'; end;
if nargin < REQ_ARG_COUNT + 2, obstacles = []; end;
if nargin < REQ_ARG_COUNT + 3 || isempty(rewards), rewards = DEFAULT_REWARDS; end;
if nargin < REQ_ARG_COUNT + 4, cfg = DEF_CFG;
else cfg = checkparams(cfg, DEF_CFG); end;

worldsize = worldsize';         % to row format
% Process goal structure
if isnumeric(goal),
    goal2 = goal; clear goal;
    goal.type = 'mixed';
    goal.pos = goal2';          % to row format
end;
if ~isfield(goal, 'rew'), goal.rew = rewards; end;
goal.pos = goal.pos';           % to row format
switch goal.type,
    case 'coop',
        % ensure there's just one goal
        goal.pos = goal.pos(1, :);
        goal.type = 1;
    case 'comp',
        goal.pos = goal.pos(1, :);
        goal.type = 2;
    case 'mixed',
        goal.type = 3;
end;

% world constants initialization
gw.name = name; gw.size = worldsize;
% action displacements in the 2D space of the grid - left, right, up, down,
% stay put
gw.actoffsets = [-1 0; 1 0; 0 -1; 0 1; 0 0]';
% grid world constants
gw.tasktype = 'episodic';
gw.type = 'robotsgw';
gw.viewtype = 'robotsgwview';


gw.grid = zeros(worldsize);

gw.acount = length(agents);
if goal.type == 2 && gw.acount ~= 2,
    error('Competitive tasks can only have two agents');
end;
% maintain a shortcut to the world state: agent positions on columns
gw.state = zeros([2 gw.acount]);
% agent ids map specifies the id for an agent on a given position in the
% state matrix
gw.aidmap = zeros([gw.acount 1]);

for i = 1:gw.acount,
    a = agents{i};
    % place agent's goal on grid if task mixed, i.e. each agent has its own
    % goal
    if goal.type == 3,
        gw.grid(goal.pos(i, 1), goal.pos(i, 2)) = 1000 + a.id;
    end;
    gw.aidmap(i) = a.id;
end;
% just one goal for cooperative and competitive settings
if goal.type < 3, gw.grid(goal.pos(1), goal.pos(2)) = 1000 + gw.aidmap(1); end; % most of times 1001

gw.goal = goal;

% if deterministic initial positions, init them now so that obstacles
% cannot be placed on top
if ~isempty(gw.initpos),
    for i = 1:gw.acount,
        gw.grid(gw.initpos(i, 1), gw.initpos(i, 2)) = agents{i}.id;
        gw.state(:, i) = gw.initpos(i, :)';
    end;
    gw.initialstate = gw.state;
else gw.state = ones(2, gw.acount);     % just a placeholder
end;

% place any specified obstacles
if isnumeric(obstacles) && ~isempty(obstacles),
    obstacles = obstacles';             % to row format
    for i = 1 : size(obstacles, 1),
        gw.grid(obstacles(i, 1), obstacles(i, 2)) = -1;
    end;
end;

gw.view = robotsgwview(gw, cfg);

if ischar(obstacles), if obstacles == 'user',
    % get obstacles interactively
    error('robotsgw() does not currently support interactive obstacles placing');
%     gw = robotsgw_obstacles(gw);
end; end;

% Init clocks
gw.k = 0; gw.kt = 0;
gw.trialsk = 0;

% returning gw

% END robotsgw()
% =======================================================================