function [gw, stateviews, actionviews, rewardviews, finished] = robotsgw_dynamics(gw, actions)
%Implements the discrete-time dynamics of the robotsgw 
%  [GW, STATEVIEWS, ACTIONVIEWS, REWARDVIEWS, FINISHED] = ROBOTSGW_DYNAMICS(GW, ACTIONS)
%  Transforms the state of the world based on the agent actions, and
%  computes the instantaneous rewards. This implementation of the robotsgw
%  will always return true and complete views over state, joint actions and
%  joint rewards.
%
%  Parameters:
%   GW          - the gridworld.
%   ACTIONS     - the joint action vector of the agents, a column vector.
%               May contain NaN elements, signifying 'noop' for agents that
%               completed their task.
%
%  Returns:
%   GW          - the altered gridworld.
%   STATEVIEWS  - the views of the agents over the state, nth agent view on
%               the nth column.
%   ACTIONVIEWS - the views of the agents over the joint action, nth agent
%               view on the nth column.
%   REWARDVIEWS - the views of the agents over the joint reward, nth agent
%               view on the nth column.
%   FINISHED    - a boolean row vector containing for each agent a flag 
%               that signals if that agents' task has finished in this trial
%               (either by success, or failure). nth agent flag on the nth
%               column.
%
%  The agents cannot move out of the grid, onto an obstacle, or in mixed
%  settings onto a goal that does not belong to them. Furthermore, if
%  several agents try to move into the same position, they will be reset to
%  their original positions (including any agents that wouldn't have hit an
%  agent if that agent wouldn't have hit something else and been reset
%  etc., recursively).
%
%  Each agent finishes when it has reached its corresponding goal. There is
%  no failure situation. In competitive worlds, the agent that reaches the
%  goal first wins, the other loses, and the trial terminates at that
%  moment.
%
%
%  See also robotsgw

%  Author:      Lucian Busoniu
%  Version:     2.0
%  History:     2.0 added coop and comp tasks handling


acount = gw.acount;

actions = actions';         % robotsgw implemented long ago, with row action standard
% find out which agents tried to act
acting = find(~isnan(actions));
still = find(actions == 5);
moved = zeros(1, acount); putback = moved; atgoal = moved;
moved(acting) = 1; moved(still) = 0;

% construct the candidate state
state = gw.state; cstate = state;
for i = acting,
    castate = cstate(:, i) + gw.actoffsets(:, actions(i));  % candidate state for agent
    % agent can only try to move onto own goal, empty space or another agent
    if all(castate >= [1;1] & castate <= gw.size'),    % check for out-of-grid
        e = gw.grid(castate(1), castate(2));                % current destination content
        % move is bad if destination is an obstacle
        % move is also bad if mixed setting and trying to move onto a goal
        % that's not your own
        if  (e < 0) || (gw.goal.type == 3 && e > 1000 && e ~= 1000 + gw.aidmap(i))
            putback(i) = 1;
        else
            cstate(:, i) = castate; atgoal(i) = e > 1000;
        end;
    else putback(i) = 1;                    % it was an attempt to move out of the grid
    end;
end;

% sw2 = all(atgoal); sw = 1;
% if sw,
% disp('After border, obstacle, and goal check:');
% cstate
% end;

% in cooperative settings, send back agents that tried to move onto goal
% independently
if gw.goal.type == 1 && ~all(atgoal),
    i = find(atgoal);
    cstate(:, i) = state(:, i); 
    putback(i) = 1;
    atgoal(i) = 0;
end;

% send back agents that try to switch positions "passing one through
% another"
movedix = find(moved & ~putback);
for ii = 1 : length(movedix) - 1,
    i = movedix(ii);
    % check whether destination cell holds an agent
    e = gw.grid(cstate(1, i), cstate(2, i));
    if e == 0 || e > 1000, continue; end;
    % if agent, test whether switching positions
    if all(cstate(:, i) == state(:, e)) && all(cstate(:, e) == state(:, i)),
        % put them back
        cstate(:, [i e]) = state(:, [i e]);
        putback([i e]) = 1;
    end;
end;

% if sw,
% disp('After switch check');
% cstate
% end;

% send any agents that bumped into each other back to their original positions
i = 1;
while i < acount,
    overlap = find(sum(abs(cstate(:, i+1:acount) - repmat(cstate(:, i), 1, acount - i))) == 0);
    if atgoal(i) || isempty(overlap),
        i = i + 1; continue;
    else
        overlap = [0 overlap] + i;                  % current agent also
        cstate(:, overlap) = state(:, overlap);     % send agents back
        putback(overlap) = 1;
        i = 1;                                      % reset search
    end;
end;

% if sw,
% disp('After bump check');
% cstate
% pause
% end;

% still agents have not been "put back" since they did not depart in the
% first place
putback(still) = 0;
% reset agents were those that were trying to move and were put back
reset = find(moved & putback);
% the ones that departed their original positions were those that weren't put back
departed = find(moved & ~putback);
% % agents that reached their goal dissappear into nothingness from the grid
% arrived = find(moved & ~putback & ~atgoal);
% allow agents to move over goals
arrived = departed;

% change the positions of the moved agents in the grid
grid = gw.grid(:);
if ~isempty(departed), grid(sub2ind(gw.size, state(1, departed), state(2, departed))) = 0; end;
if ~isempty(arrived), grid(sub2ind(gw.size, cstate(1, arrived), cstate(2, arrived))) = gw.aidmap(arrived); end;
gw.grid(:) = grid;

% promote the (now consistent) candidate state
gw.state = cstate;

% distribute rewards to agents, according to the setting
rew = 0 * moved';
switch gw.goal.type,
    case 1,         % fully cooperative
        if all(atgoal), rew(:) = gw.goal.rew(3);        % at-goal reward
        else rew(:) = gw.goal.rew(1);                   % ordinary reward
        end;
    case 2,         % fully competitive
        % if none at goal, no reward; if both at goal, a draw, so still
        % nothing
        if ~any(atgoal) || all(atgoal),
            rew(:) = 0;                    
        else
            % the one at goal receives +goal reward, the other -goal reward            
            rew(find(atgoal)) = gw.goal.rew(3);
            rew(find(~atgoal)) = -gw.goal.rew(3);
        end;
    case 3,         % mixed
        rew(:) = gw.goal.rew(1);
        rew(reset) = gw.goal.rew(2);
        rew(find(atgoal)) = gw.goal.rew(3);
end;

% generate views for all agents
stateviews = repmat(cstate(:), 1, acount);
actionviews = repmat(actions', 1, acount);      % back to column format
rewardviews = repmat(rew, 1, acount);

% agents are finished if they were finished before, they reached the goal,
% or in a fully comp case, if any reached the goal
if gw.goal.type == 2 && any(atgoal),
    finished = 1 + 0 * atgoal;
else finished = atgoal | isnan(actions);
end;

% update view
gw = robotsgw_display(gw);

% increment clock
gw.k = gw.k + 1;        % general clock
gw.kt = gw.kt + 1;      % within trial

% returning gw, stateviews, actionviews, rewardviews, finished

% END robotsgw_advance() ==================================================