function [gw, stateviews] = robotsgw_reset(gw, restype)
%Resets the robotsgw to its initial state and reports this state
%  [GW, STATEVIEWS] = robotsgw_RESET(GW, [RESTYPE])
%  Resets the robotsgw to its initial state, after a trial has ended.
%  Returns the state of the world as viewed by all the agents within.
%  All agents will see the true state of the robotsgw.
%
%  Parameters:
%   GW          - the robotsgw
%   RESTYPE     - (optional) the reset type, 0 soft, 1 hard. Default 0.
%  Returns:
%   GW          - the reset robotsgw
%   STATEVIEWS  - the state views, one agent view per column
%
%
%  See also robotsgw, robotsgw_advance

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


if nargin < 2, restype = 0; end;

% reset goal location(s), as they might have been overwritten by agents
if gw.goal.type < 3,    % coop or comp
    gw.grid(gw.goal.pos(1), gw.goal.pos(2)) = 1000 + gw.aidmap(1);
else                    % mixed 
    for i = 1:gw.acount,
        gw.grid(gw.goal.pos(i, 1), gw.goal.pos(i, 2)) = 1000 + gw.aidmap(i);
    end;
end;

% flatten grid, remove any agents and place them in their start positions
grid = gw.grid(:);
grid(find(grid > 0 & grid < 1000)) = 0;

prevstate = gw.state;
if ~isempty(gw.initpos),
    % reset to deterministic initial positions
    grid(sub2ind(gw.size, gw.initialstate(1, :), gw.initialstate(2, :))) = gw.aidmap;
    gw.state = gw.initialstate;
else
    % init random positions
    for i = 1:gw.acount,
        grid = reshape(grid, gw.size);
        pos = ceil(rand(1, 2) .* gw.size);
        while grid(pos(1), pos(2)) ~= 0, pos = ceil(rand(1, 2) .* gw.size);
        end;
        grid(pos(1), pos(2)) = gw.aidmap(i);
        gw.state(:, i) = pos';
    end;
end;


gw.grid(:) = grid;
gw = robotsgw_display(gw);

stateviews = repmat(gw.state(:), 1, gw.acount);

% increment trials clock
if restype == 0,
    gw.trialsk = gw.trialsk + 1;
elseif restype > 0,
    gw.k = 0;
    gw.trialsk = 0;
end;
gw.kt = 0;      % within-trial clock

% returning gw, stateviews
% END robotsgw_state() ===================================================