function [gw, stateviews, actionviews, rewardviews, finished] = rrgw_dynamics(gw, actions)
%Implements the discrete-time dynamics of the rrgw 
%  [GW, STATEVIEWS, ACTIONVIEWS, REWARDVIEWS, FINISHED] = RRGW_DYNAMICS(GW, ACTIONS)
%  Transforms the state of the world based on the agent actions, and
%  computes the instantaneous rewards. This implementation of the rrgw
%  will always return true and complete views over state, joint actions and
%  joint rewards.
%
%  Parameters:
%   GW          - the gridworld.
%   ACTIONS     - the joint action vector of the agents, a column vector.
%               May contain NaN elements, signifying 'noop' for agents that
%               completed their task.
%
%  Returns:
%   GW          - the altered gridworld.
%   STATEVIEWS  - the views of the agents over the state, nth agent view on
%               the nth column.
%   ACTIONVIEWS - the views of the agents over the joint action, nth agent
%               view on the nth column.
%   REWARDVIEWS - the views of the agents over the joint reward, nth agent
%               view on the nth column.
%   FINISHED    - a boolean row vector containing for each agent a flag 
%               that signals if that agents' task has finished in this trial
%               (either by success, or failure). nth agent flag on the nth
%               column.
%
%  The agents cannot move out of the grid, onto an obstacle, or in mixed
%  settings onto a goal that does not belong to them. Furthermore, if
%  several agents try to move into the same position, they will be reset to
%  their original positions (including any agents that wouldn't have hit an
%  agent if that agent wouldn't have hit something else and been reset
%  etc., recursively).
%
%  Each agent finishes when it has reached its corresponding goal. There is
%  no failure situation. In competitive worlds, the agent that reaches the
%  goal first wins, the other loses, and the trial terminates at that
%  moment.
%
%
%  See also rrgw

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:     


% Tested functionality:
%   crossing between agents CHECK
%   collisions of agent with agent, obstacle, pinned conglomerate CHECK
%   pinned agent, ripping agent CHECK
%   collisions of conglomerates with various parts: CHECK
%   MIXED SETTING (testing only for two agents 
%                   => coll with pinned and the like not tested in this setting):
%   valid and invalid switches btw conglomerates: CHECK
%   collisions between conglomerates and: agents, other conglomerates: CHECK
%   collision with still / home conglomerate CHECK
% compute rewards       CHECK
% test rewards          CHECK (using scenarios above)

agn = gw.agn; targn = gw.targn; objn = agn+targn;

actions = actions';         % row action standard

% state shorctut
state = gw.state; 

% ------------------------
% INIT flags and other stuff
moving_flag         = false(1, agn);
pulling_flag        = moving_flag;
ripping_flag        = moving_flag;
pinned_flag         = moving_flag;
sittinglocked_flag  = moving_flag;

% target-related variables
tmoving_flag        = false(1, targn);
target_dirs         = 0 * tmoving_flag;     % convert to numerical, from logical flag

% initially all objects are sitting in their own "conglomerates"
conglomerates = num2cell(1:objn);

% ============================
% Collision check and stuff preparation... painful indeed

% find out which agents tried to act
acting_flag   = ~isnan(actions);
still_flag    = actions == 5;
% find out which agents are actually atempting to move
moving_flag(acting_flag) = 1; moving_flag(still_flag) = 0;


% construct flags for FIXED OBJECTS, against which collisions may easily be checked
% obstacles are always fixed
fixedpos = (gw.grid == -1);         

% note that currently, the flw are possibly redundant checks:
%   MOVING OBJECTS collision check should catch collisions of moving objects with static
%       objects as well
%   whereas here, we set fixed_pos for static objects and check them before that
% if removing redundancy, remove the fixed flags settings for objects below
%   (targets, still agents etc)

% finished or still agents are fixed
for i = find(~moving_flag), fixedpos(state(1, i), state(2, i)) = 1; end;

% check targets status
for i = 1:gw.targn,
    % which agents are locked to this target
    lockedagents = gw.lockflags(gw.targets{i}.lockindices);
    if ~all(lockedagents),
        % all agents which are locked to partially locked targets are pinned
        pinned_flag( lockedagents(find(lockedagents)) ) = 1;
        % and that target is pinned, so mark it as FIXED and forget about it
        fixedpos(state(1, agn + i), state(2, agn + i)) = 1;
        continue;
    end;
    % otherwise, check if agents are using coordinated directions
    act = actions(lockedagents);
    if any(diff(act)),      % agents are ripping the target!
        ripping_flag(lockedagents) = 1;
    elseif all(act == 5 | isnan(act)),        % agents are nicely standing still (or finished at homebase)
        sittinglocked_flag(lockedagents) = 1;
        % this target is a FIXED OBJECT
        fixedpos(state(1, agn + i), state(2, agn + i)) = 1;
    else                    % agents are pulling nicely
        pulling_flag(lockedagents) = 1;
        % so, the target is attempting to move
        tmoving_flag(i) = 1;
        % in this direction:
        target_dirs(i) = act(1);
        % remember the conglomerate formed by this target and its locked agents
        % for eventual putting back
        conglomerate = [agn + i, lockedagents];
        % place the conglomerate at every index of its components
        for ii = 1:length(conglomerate),
            conglomerates{conglomerate(ii)} = conglomerate;
        end;
    end;
    
    % note that actually, still agents + targets are also conglomerates, but for the purposes
    % of the algorithm below they are treated as still objects, so they can be handled
    % component-wise
end;

% joint, object variables (caps)
% complete moving flags array: agents + targets
MOVING_flag         = [moving_flag tmoving_flag];
% complete flags for objects which are put back
PUTBACK_flag        = false(size(MOVING_flag));
% complete directions array
DIRS                = [actions target_dirs];
% home free
HOME_flag           = PUTBACK_flag;

% candidate state
candidate = state;

% =========================
% PINNED OR RIPPING AGENTS

% first of all, put back agents which are pinned or ripping
% there is no danger for these agents to be in a valid conglomerate. so it's fine to just set
% their own putback flag
PUTBACK_flag( moving_flag & (pinned_flag | ripping_flag) ) = 1;
% no sense checking them again
CRTMOV_ix = find(MOVING_flag & ~PUTBACK_flag);

% =======================
% COLLISIONS WITH FIXED OBJECTS
% then check for collisions with fixed objects. Order of putting back here does not matter,
% as we are only checking againts fixed objects
for i = CRTMOV_ix,
    if PUTBACK_flag(i), continue; end;
    candpos = candidate(:, i) + gw.actoffsets(:, DIRS(i));  % candidate position for agent
    % deny move to objects which are colliding with fixed objects
    % or attempt to exit the grid
    if (any(candpos < [1;1] | candpos > gw.cfg.size)) || fixedpos(candpos(1), candpos(2)),
        PUTBACK_flag(conglomerates{i}) = 1;
        candidate(:, conglomerates{i}) = state(:, conglomerates{i});
    else        % validate, for now
        candidate(:, i) = candpos;
    end;
end;


% =======================
% POSITION SWITCHES
% send back objects that (conglomerates within which objects) tried to switch positions
% "passing one through another"
CRTMOV_ix = find(MOVING_flag & ~PUTBACK_flag);       % recompute CRTMOV_ix
for ii = 1 : length(CRTMOV_ix) - 1,
    i = CRTMOV_ix(ii);
    % skip the rest if already put back
    if PUTBACK_flag(i), continue; end;
    % check whether destination cell is a movable object
    e = gw.grid(candidate(1, i), candidate(2, i));
    if e == 0 || (e > 100 && e <= 1000), continue; end;
    e = mod(e, 1000) + (e > 1000) * agn;        % actual object index
    % moving object; test whether switching positions with original
    if all(candidate(:, i) == state(:, e)) && all(candidate(:, e) == state(:, i)),
        % put the (possible conglomerates) back
        candidate(:, conglomerates{i}) = state(:, conglomerates{i});
        candidate(:, conglomerates{e}) = state(:, conglomerates{e});
        PUTBACK_flag(conglomerates{i}) = 1;
        PUTBACK_flag(conglomerates{e}) = 1;
    end;
end;


% =======================
% COLLISIONS BETWEEN MOVING OBJECTS (CONGLOMERATES) AND OTHER, FIXED OR MOVING, OBJECTS
% send any ojbects (conglomerates) that bumped into each other back to their original positions
i = 1;

while i < objn,
    overlap = find(sum(abs(candidate(:, (i + 1):objn) - repmat(candidate(:, i), 1, objn - i))) == 0);
    if isempty(overlap),
        i = i + 1; continue;
    else
        overlap = [0 overlap] + i;                      % current agent also
        for ii = 1:length(overlap),
            candidate(:, conglomerates{overlap(ii)}) = state(:, conglomerates{overlap(ii)});
            PUTBACK_flag(conglomerates{overlap(ii)}) = 1;
        end;
        i = 1;                                          % reset search
    end;
end;

% =======================
% UPDATE GRID, UPDATE LOCKS, CHECK FOR GOAL
locksview = gw.lockflags;

% process agents
putback_flag = PUTBACK_flag(1:agn);
moved_flag = MOVING_flag(1:agn) & ~putback_flag;
locked_flag = false(size(putback_flag));
for i = find(moved_flag),
    dest = candidate(:, i);
    % 1. CHECK IF AGENT REACHED LOCK POINT, lock them if so
    % check whether destination doesn't match a lock position
    lockix = find(gw.lockpos(1, :) == dest(1) & gw.lockpos(2, :) == dest(2));
    if ~isempty(lockix),
        gw.lockflags(lockix) = i;
        gw.lockpos(:, lockix) = 1e6;       % take out of further comparisons
        % append to locks view
        locksview(i) = lockix;
        locked_flag(i) = 1;
    end;
    
    % 2. change position on grid
    gw.grid(gw.grid == i) = 0; gw.grid(dest(1), dest(2)) = i;
end;

% process targets
tmoved_flag = MOVING_flag(agn+1:end) & ~PUTBACK_flag(agn+1:end);
for i = find(tmoved_flag),
    % 1. CHECK IF TARGET REACHED HOMEBASE, mark objects as homefree
    dest = candidate(:, agn+i);
    home = gw.cfg.homebase(1, 1) <= dest(1) && dest(1) <= gw.cfg.homebase(1, 2) ...
        && gw.cfg.homebase(2, 1) <= dest(2) && dest(2) <= gw.cfg.homebase(2, 2);
    if home, HOME_flag(conglomerates{agn+i}) = 1; end;
    
    % 2. change position on grid
    gw.grid(gw.grid == 1000+i) = 0; gw.grid(dest(1), dest(2)) = 1000+i;
end;

% flags postprocessing
% collided agents
home_flag = HOME_flag(1:agn);
collided_flag = putback_flag & ~still_flag & ~ripping_flag & ~pinned_flag;


% promote the (now consistent) candidate state
gw.state = candidate;


% =============================
% COMPUTE REWARDS
% distribute rewards to agents, according to the setting
rewcfg = gw.cfg.rew;
rew = zeros(agn, 1) + rewcfg.step;
% in inverse order of priority: those above can overwrite those below
% most are mutually exclusive
rew(pulling_flag) = rewcfg.pull;
rew(sittinglocked_flag) = rewcfg.sit;
rew(collided_flag) = rewcfg.coll;       % collision takes priority over pull
rew(ripping_flag) = rewcfg.rip;         % ripping is typically worse (although coll computed above is exclusive)
rew(locked_flag) = rewcfg.lock;         % positive rewards take priority over rest
rew(home_flag) = rewcfg.home;
if (gw.cfg.tasktype == 1),
    % if cooperative, agents cannot have different rewards; a minimum (worst-case) is computed
    % across the agents. s.t. e.g. one agent collides, all get the penalty, and only when all
    % agents are at homebase is the homebase reward distributed
    rew(:) = min(rew);
    % however, lock events should be noted
    if any(locked_flag), rew(:) = rewcfg.lock; end;
end;


% generate views for all agents; locks below positions, and displaced with 1 (to start from 1)
stateviews = repmat(reshape([gw.state(:, 1:agn); locksview+1], [], 1), 1, agn);
actionviews = repmat(actions', 1, agn);      % back to column format
rewardviews = repmat(rew, 1, agn);

% agents are finished if finished before or homefree
finished = home_flag | isnan(actions);

% update view
gw = rrgw_display(gw);

% increment clock
gw.k = gw.k + 1;        % general clock
gw.kt = gw.kt + 1;      % within trial

% returning gw, stateviews, actionviews, rewardviews, finished

% END rrgw_advance() ==================================================