function [info] = world_info(w)
%Template. Generates a data structure with information about the world.
%  [INFO] = WORLD_INFO(W)
%  Returns information about the world that could be useful to the agents.
%
%  This function is a template and performs no operations.
%  A new world type must implement this function under the name
%  <world_type>_indexmaps. This functions' input and output argument(s)
%  must conform to the specifications below.
% 
%  Parameters:
%   W           - the world
%  Returns:
%   INFO        - the world information structure
%
%  INFO must contain the following fields:
%
%  - 'imaps' - an index maps matrix that specifies where in the world
%  state vector each agent finds its own state and the common state of the
%  world (i.e. state  elements independent of other agents), and where on
%  the joint action and reward vectors the agent finds its own action and 
%  reward. Agent n finds its index map on the nth column of the IMAPS matrix, 
%  which has the structure:
%     [AS1 AS2 CWS1 CWS2 AIX1 AIX2 RIX]', where:
%       - STATE(AS1:AS2) represents the agent's state.
%       - STATE(CWS1:CWS2) represents the common world state.
%       - ACTION(AIX1:AIX2) is the agent's action in the joint action
%       vector.
%       - RIX is the agent's index in the joint reward vector.
%
%   - 'stattespacesize' A matrix containing the sizes of the state 
%   components, one row for each state component. The first column
%   specifies the type of component, 1 for discrete, 2 for continuous
%   For continuous components, columns 2-3 specify the ends of the
%   interval inside which the state varies. The 4th column is a binary flag
%   specifying if the state component "rolls over" from one interval end to
%   the other. For components that do not roll over, infinite bounds are
%   allowed, given that they are supported by the agent algorithm.
%   Discrete components have the cardinality of the component on the 2nd
%   column. The rest of the column is not taken into account.
%
%   - 'actionspacesize' the size of the joint action space. Formed in
%   the same way as the above.
%
%   - 'statenames' a cell array, each element containing a unique string
%   name for the corresponding state component.
%
%   - 'actionnames' a cell array, each element containing a unique string
%   name for the corresponding action component.
%
%  See below for an example of how to form these arrays.
%
%  The size descriptors must have the same length, thus, if at least one of
%  the states or actions is continuous, all state and action size
%  descriptors must have 4 columns.
%
%  Naturally, the order of the state and action components in the arguments
%  must be consistent, such that e.g., the agent finds the size of its own
%  state space at the position specified in its own index map.
%
%  See also world, agent_initlearn

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


% An example of forming the info structure

% info.statespacesize = ...
%     [1      10      NaN     NaN         % one discrete component with 10 discrete values
%      2      0       2*pi    1           % one continuous component in [0, 2pi) that rolls over
%      2      -inf    inf     0];         % one continuous component in Reals
%  
% info.statenames = {'level', 'theta', 'x'};
%      
% info.actionspacesize = ...
%     [2      -10     10      0];         % one continuous component in [-10 10] that does not roll over
% 
% info.actionnames = {'torque'};
    

% returning iinfo
% END world_info() ========================================================
