function [a, action] = randomexplore(a, action, state, actions, rewards, params)
%Implements non-exploitive random exploration.
%  [A, ACTION] = RANDOMEXPLORE(A, ACTION, STATE, ACTIONS, REWARDS, PARAMS)
%  Implements non-exploitive random exploration with probability epsilon.
%  This is the most basic (yet the most commonly used) exploration
%  strategy. Can be used with any learning/policy functions combination, if 
%  the required parameters listed below are fed to the learning
%  initialization function.
%
%  Supports discrete actions, with 1 action variable only.
%
%  Required values on the agent learning parameters:
%   epsilon         - the exploration probability
%   epsilondecay    - the exploration probability decay ratio. Required
%                   in non-episodic environments. If specified, epsilon
%                   will decay with this ratio at each iteration,
%                   otherwise, in episodic environments, it will decay with
%                   the trials count.
%  Required values on extra parameters:
%   newtrial        - (in episodic environments) whether a new trial is
%                   beginning
%
%
%  See also act, explore

%  Author:      Lucian Busoniu
%  Version:     1.0
%  History:


if isfield(params, 'mode'),
    switch params.mode,
        case 'init',
            a.learnparam.epsilon0 = a.learnparam.epsilon;
            return;
        case 'reset',
            if isfield(params, 'weight'),
                w = params.weight;
            else w = 1;
            end;
            a.learnparam.epsilon = w * a.learnparam.epsilon0;
            return;
    end;
end;

if rand < a.learnparam.epsilon, action = ceil(rand * a.actionscount); a.exploring = 1;
else a.exploring = 0;
end;

if isfield(a.learnparam, 'epsilondecay'),
    a.learnparam.epsilon = a.learnparam.epsilon * a.learnparam.epsilondecay;
elseif params.newtrial,
    if isfield(a.learnparam, 'epsilontrialdecay') && a.trialsk > 1,
        a.learnparam.epsilon = a.learnparam.epsilon * a.learnparam.epsilontrialdecay;
    else
        a.learnparam.epsilon = a.learnparam.epsilon / a.trialsk;
    end;
end;
    
% returning a, action
% END randomexplore() =====================================================